#
# Copyright 2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

import trafaret as t

from datarobot.errors import ClientError

try:
    # Literal is only available from typing in Python 3.8+. For 3.7 you need to use typing_extensions
    from typing import Literal
except ImportError:
    from typing_extensions import Literal

from datarobot._compat import Int, String
from datarobot.enums import (
    AUTOPILOT_MODE,
    CV_METHOD,
    TARGET_TYPE,
    TIME_UNITS,
    TREAT_AS_EXPONENTIAL,
    VALIDATION_TYPE,
)
from datarobot.helpers import (
    _class_mapping_aggregation_settings_converter,
    AdvancedOptions,
    ClassMappingAggregationSettings,
)
from datarobot.helpers.partitioning_methods import (
    _feature_settings_converter,
    BacktestSpecification,
    FeatureSettings,
)
from datarobot.models.api_object import APIObject
from datarobot.models.feature import Feature
from datarobot.utils import from_api, parse_time


class RelationshipGraph:
    """
    A graph showing related datasets. linkage_keys is a list of dataset ids that are related.
    """

    _converter = t.Dict(
        {
            t.Key("id", optional=True): String(),
            t.Key("linkage_keys", optional=True): t.List(String()),
        }
    )
    schema = _converter

    def __init__(self, id: str = None, linkage_keys: List[str] = None):
        self.id = id
        self.linkage_keys = linkage_keys

    def collect_payload(self) -> Dict[str, Any]:
        return {"id": self.id, "linkage_keys": self.linkage_keys}


class FeatureEngineeringDataset:
    """
    Dataset related to other datasets when using automated feature engineering.
    """

    _converter = t.Dict({t.Key("catalog_id", optional=True): String()})
    schema = _converter

    def __init__(self, catalog_id: str = None):
        self.catalog_id = catalog_id

    def collect_payload(self) -> Dict[str, Any]:
        return {"catalog_id": self.catalog_id}


class FeatureEngineeringOptions:
    """
    Options used for automated feature engineering.
    """

    _converter = t.Dict(
        {
            t.Key("related_graphs", optional=True): t.List(RelationshipGraph.schema),
            t.Key("related_datasets", optional=True): t.List(FeatureEngineeringDataset.schema),
            t.Key("suggested_graph_ids", optional=True): t.List(String()),
            t.Key("is_snowflake_writable", optional=True): t.Bool(),
        }
    )
    schema = _converter

    def __init__(
        self,
        related_graphs: List[RelationshipGraph] = None,
        related_datasets: List[FeatureEngineeringDataset] = None,
        suggested_graph_ids: List[str] = None,
        is_snowflake_writable: bool = None,
    ):
        if related_graphs and isinstance(related_graphs[0], dict):
            self.related_graphs = [RelationshipGraph(**graph) for graph in related_graphs]
        else:
            self.related_graphs = related_graphs
        if related_datasets and isinstance(related_datasets[0], dict):
            self.related_datasets = [
                FeatureEngineeringDataset(**dataset) for dataset in related_datasets
            ]
        else:
            self.related_datasets = related_datasets
        self.suggested_graph_ids = suggested_graph_ids
        self.is_snowflake_writable = is_snowflake_writable

    def collect_payload(self) -> Dict[str, Any]:
        return {
            "related_graphs": [graph.collect_payload() for graph in self.related_graphs]
            if self.related_graphs
            else None,
            "related_datasets": [dataset.collect_payload() for dataset in self.related_datasets]
            if self.related_datasets
            else None,
            "suggested_graph_ids": self.suggested_graph_ids,
            "is_snowflake_writable": self.is_snowflake_writable,
        }


class Duration:
    """
    Duration information for a holdout.
    """

    _converter = t.Dict(
        {
            t.Key("max_length", optional=True): Int(),
            t.Key("min_length", optional=True): Int(),
            t.Key("allow_blank", optional=True): t.Bool(),
            t.Key("convert_to_timedelta", optional=True): t.Bool(),
        }
    )

    def __init__(
        self,
        max_length: int = None,
        min_length: int = None,
        allow_blank: bool = None,
        convert_to_timedelta: bool = None,
    ):
        self.max_length = max_length
        self.min_length = min_length
        self.allow_blank = allow_blank
        self.convert_to_timedelta = convert_to_timedelta

    def collect_payload(self) -> Dict[str, Any]:
        return {
            "max_length": self.max_length,
            "min_length": self.min_length,
            "allow_blank": self.allow_blank,
            "convert_to_timedelta": self.convert_to_timedelta,
        }


class Periodicity:
    """
    For time series projects only. Includes time step and time unit information for a given modeling period.
    """

    _converter = t.Dict(
        {t.Key("time_steps", optional=True): Int(), t.Key("time_unit", optional=True): String()}
    )

    def __init__(self, time_steps: int = None, time_unit: Union[TIME_UNITS, Literal["ROW"]] = None):
        self.time_steps = time_steps
        self.time_unit = time_unit

    def collect_payload(self) -> Dict[str, Any]:
        return {
            "time_steps": self.time_steps,
            "time_unit": self.time_unit,
        }


_cv_method_map = {
    "UserCV": CV_METHOD.USER,
    "DatetimeCV": CV_METHOD.DATETIME,
    "GroupCV": CV_METHOD.GROUP,
    "RandomCV": CV_METHOD.RANDOM,
    "StratifiedCV": CV_METHOD.STRATIFIED,
}


def cv_method_converter(dr_cv_method: str) -> str:
    """
    Parameters
    ----------
    dr_cv_method : str : The CV method coming from DataRobot's project options endpoint

    Returns
    -------
    The string value from :enum:`CV_METHOD <datarobot.enums.CV_METHOD>`
    """
    return _cv_method_map[dr_cv_method]


if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    class PartitioningWarnings(TypedDict):
        backtest_index: Optional[int]
        partition: str
        warnings: List[str]

    class ExtendedWarning(TypedDict):
        title: str
        message: str
        type: str

    class PartitioningExtendedWarnings(TypedDict):
        backtest_index: Optional[int]
        partition: str
        warnings: List[ExtendedWarning]


class ProjectOptions(AdvancedOptions, APIObject):
    """
    All available options set for a project, including time series specific settings.

    Parameters
    ----------
    project_id : str
        The project this belongs to
    target : str
        Column indicating the target of modeling
    target_type : str
        The user defined target_type, enum TargetType ['Multiclass', 'Binary', 'Regression']
    initial_mode : int, optional
        Choices: [AUTO=0, SEMI=1, MANUAL=2]
        (see common.services.autopilot for the constants)
    initial_num_workers : int, optionalma
        The amount of workers that will be used during modeling
    initial_gpu_workers : int, optional
        The amount of gpu workers that will be used during modeling
    auto_start : boolean, optional
        Determines whether EDA2 and modeling will begin as  soon as EDA1 is finished.
        Defaults to False.
    is_dirty : boolean, optional
        Whether this set of project options has been altered by the user (or is the default)
    metric : str, optional
        The metric.
    weights : str, optional
        Column used for weight during modeling
    blueprint_threshold : int, optional
        Upper bound on running time of models in hours
    blend_best_models : bool, optional
        Blend best models during Autopulot run
    min_secondary_validation_model_count : int, optional
        Number of models to compute secondary validation scores (cv/all backtests)
        if over Autopilot default. To be used in when row count is larger
        than autopilot cv/backtest threshold.
    prepare_model_for_deployment : bool, optional
        Prepare model for deployment during Autopilot run. The preparation includes creating
        reduced feature list models, retraining best model on higher sample size,
        computing insights and assigning "RECOMMENDED FOR DEPLOYMENT" label.
    scoring_code_only : bool, optional
        Exclude blueprints that cannot be converted to scorable code
    shap_only_mode  : bool, optional
        Keep only models that support SHAP values during Autopilot run. Use SHAP based insights
        wherever possible. Defaults to False.
    consider_blenders_in_recommendation: bool, optional
        Whether to allow blenders to be considered when a model is chosen to be prepared for
        deployment during an Autopilot run.
    run_leakage_removed_feature_list: bool, optional
        Run Autopilot on feature list with target leakage removed
    seed : int, optional
        The seed for the randomness algorithm during RandomCV
    smart_downsample_enabled : boolean, optional
        Whether smart downsampling is enabled, defaults to False
    smart_downsample_rate : float, optional
        Percentage describing new size of majority class after downsampling
    validation_type : str, optional
        Type of validation, `CV` for Cross Validation or `TVH` for Train-Validation-Holdout
    cv_method : str, optional
        Method of validation, `RandomCV`, `StratifiedCV`, `UserCV`, `GroupCV`, `DateCV`,
        'DatetimeCV'
    reps : int, optional
        Relevant only for Cross Validation, Integer specifying number of validation folds
    validation_pct : float, optional
        Relevant only for TVH, percentage indicating size of validation partition
    holdout_pct : float, optional
        Percentage indicating size of holdout partition
    user_partition_col : float, optional
        Column used for user partitioning
    cv_holdout_level : float, optional
        Value from partition column used for holdout level when CV is chosen, mutually exclusive
        with the next three.
    training_level : depends on column type, optional
        Value from user partition column that corresponds to training partition
    validation_level : depends on column type, optional
        Value from user partition column that corresponds to validation partition
    holdout_level : depends on column type, optional
        Value from user partition column that corresponds to holdout partition
    partition_key_cols : str, optional
        Column used for group partitioning
    feature_list_id : str, optional
        Indicates what featurelist to use
    accuracy_optimized_mb : boolean, optional
        Whether accuracy-optimized metablueprint is enabled
    response_cap : float, optional
        Cap value for responses
    error_message : str, null, optional
        Indicates if there's an error present on Aim UI
    datetime_partition_column: str, null, optional
        The date column that will be used as a datetime partition column
    validation_duration: str, null, optional
        The default validation duration for all backtests.
    disable_holdout: boolean, optional
        Indicates whether datetime partitioning should skip allocating the main holdout fold.
    is_holdout_modified: boolean, optional
        Indicates whether holdout start/end date settings have been modified by user.
    feature_discovery_supervised_feature_reduction: boolean, optional
        Used for feature discovery projects.
        Whether to enable supervised feature reduction step.
    holdout_start_date: str, null, optional
        The holdout start date of the main fold
    holdout_end_date: str, null, optional
        The holdout end date of the main fold
    holdout_duration: str, null, optional
        The holdout duration of the main fold.
    gap_duration: str, null, optional
        The duration of the gap between the training and the holdout of the main fold.
    number_of_backtests: int, null, optional
        The number of backtests to use
    autopilot_data_selection_method: str, null, optional
        Whether models created via the autopilot will use
        "rowCount" or "duration" as their dataSelectionMethod.
    autopilot_data_sampling_method: str, optional
        Sampling method for selecting training data subsamples on autopilot iterations
    backtests: array, BacktestSpecification, optional
        An array specifying the format of the backtests, as detailed below.
        If any backtest is left unspecified, default values will be supplied.
    use_time_series : bool, optional
        Whether to use time-series with datetime partitioning
    default_to_a_priori : bool, optional
        Whether to default to treating all features as a priori
    default_to_do_not_derive : bool, optional
        Whether to default to treating all features as "do-not-derive"
    feature_settings : list of dict, optional
        List of feature settings
    feature_derivation_window_start : int, optional
        Used for time-series.  The offset (into the past) to the start of the feature derivation
        window.  Expected to be non-positive.
    feature_derivation_window_end : int,  optional
        Used for time-series.  The offset (into the past) to the start of the feature derivation
        window.  Expected to be non-positive.
    windows_basis_unit: string, optional
        Used for time series.  Indicates which unit is basis for feature derivation window and
        forecast window. Can be either detected time unit (one of AllowedTimeUnitsFEAR.ALL) or
        "ROW". If omitted, the default value is detected time unit.
    forecast_window_start : int,  optional
        Used for time-series.  The offset (into the future) to the start of the forecast
        window.  Expected to be non-negative.
    forecast_window_end : int,  optional
        Used for time-series.  The offset (into the future) to the end of the forecast
        window.  Expected to be non-negative.
    external_predictions : list of str, optional
        List of columns with external predictions
    events_count : str, optional
        Column with events count
    multiseries_id_columns : list of str, optional
        Columns with multiseries ids, aka cross sectional dimension
    calendar_id : ObjectId, optional
        The id of the calendar in use in the project.
    calendar_name : str, optional
        The name of the calendar for time-series projects.
    unsupervised_mode : bool, optional
        If True, unsupervised project (without target) will be created. False if omitted.
    unsupervised_type: str, optional
        The type of unsupervised project: anomaly or clustering
    feature_engineering_graphs : array of dict, optional
        An array specifying the Feature Engineering Graphs to be used
    relationships_configuration_id : string, optional
        Identifier of Relationships Configuration to be used
    primary_location_column : The primary geospatial column selected by the user
    autopilot_with_feature_discovery: bool, optional
        if true, autopilot will run on a feature list that includes features found via auto feature
        discovery
    date_removal: bool, optional
        if true, enable creating additional feature lists without dates
        (does not apply to time-aware projects).
    feature_engineering_options : array of FeatureEngineeringOptions, optional
        An array specifying the options for automated feature engineering
    feature_engineering_prediction_point: str, null, optional
        For time-aware Feature Engineering, this parameter specifies the column from the
        primary dataset to use as the prediction point.
    allowed_pairwise_interaction_groups: list of list of str
        For GAM models - specify groups of columns for which pairwise interactions will be allowed.
        E.g. if set to [["A", "B", "C"], ["C", "D"]] then GAM models will allow interactions
        between columns AxB, BxC, AxC, CxD. All others (AxD, BxD) will not be considered.
        If not specified - all possible interactions will be considered by model.
    allowed_pairwise_interaction_groups_filename: str
        Filename that was used to upload allowed_pairwise_interaction_groups.
        Necessary for persistence of UI/UX when you specify that parameter via file.
    model_splits: int, optional
        set the cap on the number of jobs used when building models to control number of jobs in
        the queue; also allows less downsampling for use of more post-processed data
    quantile_level: float, optional
        Only applicable to regression projects using the Quantile Loss metric.
        User-specified quantile level at which to optimize quantile loss; between 0 and 1.
    external_time_series_baseline_dataset_id: str, optional
        Catalog dataset id for external prediction dataset used to generate new metrics with
        external predictions as baseline
    external_time_series_baseline_dataset_name: str, optional
        The name of the time series baseline dataset name for the project
    preferable_target_value : str, optional
        a target value that should be treated as a positive outcome for the prediction.
        For example if we want to check gender discrimination for giving a loan
        and our target named ``is_bad``, then the positive outcome for the prediction
        would be ``No``, which means that the loan is good and that's what we treat
        as a preferable result for the loaner.
        Used and required for Bias & Fairness calculation.
    protected_features : list of str, optional
        list of the project features that are consumed as protected for calculated bias metrics.
        Only categorical features are allowed.
        Used and required for Bias & Fairness calculation.
    fairness_metrics_set : str or None, optional
        metric to use for bias calculation. One of common.enum.BiasMetric.
        Used and required for Bias & Fairness calculation.
    fairness_threshold : float or None, optional
        threshold value of the fairness metric. Can be in a range of ``[0:1]``.
        If the actual metric value is below the threshold, the user will be notified about that.
    bias_mitigation_feature_name : str of None, optional
        The feature from protected features that will be used in a bias mitigation task to
        mitigate bias
    bias_mitigation_technique : str or None, optional
        One of BiasMitigationTechnique.
        The technique by which we'll mitigate bias, which will inform which bias mitigation task
        we insert into blueprints and how
    include_bias_mitigation_feature_as_predictor_variable : bool or None, optional
        Whether we should also use the mitigation feature as in input to the modeler just like
        any other categorical used for training, i.e. do we want the model to "train on" this
        feature in addition to using it for bias mitigation
    min_clusters: int, optional
        The minimum number of clusters allowed when training clustering models.
    max_clusters: int, optional
        The maximum number of clusters allowed when training clustering models
    segmentation_id_column: string, optional
        The segmentation column name or automated segmentation column name specified for the
        project through the UI
    segmentation_task_id: string, optional
        The segmentation task id selected for segmenting the project
    segmentation_model_package_id: string, optional
        The model registry package id selected used for segmentation
    segmentation_model_package_name: string, optional
        The name of the model contained in the segmentation model registry package
    segments_count: int, optional
        The segments count for the segmented project
    allow_partial_history_time_series_predictions: bool, optional
        Whether to allow time series predictions with partial history
    exponentially_weighted_moving_alpha: float, optional
        Discount factor (alpha) used for exponentially weighted moving features
    autopilot_cluster_list : list of int
        A list of integers where each value will be used as the number of
        clusters in Autopilot model(s) for unsupervised clustering projects.
    use_project_settings : bool
        Whether to use project settings
        (i.e. backtests configuration has been modified by the user).
    use_supervised_feature_reduction: bool, optional
        Whether to use time series supervised feature reduction.
    class_mapping_aggregation_settings: ClassMappingAggregationSettings, optional
        For multiclass additional settings can be specified to control aggregation of target
        values in final classes.
    class_mapping_aggregation_settings_enabled: bool, optional
        Is the class mapping aggregation settings section enabled or not in UI
    datetime_partitioning_id: str, optional
        Id of the datetime partitioning to use
    use_gpu: bool, optional
        Whether to allow GPU usage for the project
    """

    _backtest_specification_converter = t.Dict(
        {
            t.Key("gap_duration", optional=True): String(),
            t.Key("index", optional=True): Int(),
            t.Key("primary_training_end_date", optional=True): parse_time,
            t.Key("primary_training_start_date", optional=True): parse_time,
            t.Key("validation_duration", optional=True): String(),
            t.Key("validation_end_date", optional=True): parse_time,
            t.Key("validation_start_date", optional=True): parse_time,
        }
    ).ignore_extra("is_modified")

    _partitioning_warnings_converter = t.Dict(
        {
            t.Key("backtest_index"): t.Int(gte=0) | t.Null,  # None for holdout
            t.Key("partition"): t.String,
            t.Key("warnings"): t.List(t.String),
        }
    ).allow_extra("*")

    _partitioning_extended_warnings_converter = t.Dict(
        {
            t.Key("backtest_index"): t.Or(t.Int(gte=0), t.Null),  # None for holdout
            t.Key("partition"): t.String,
            t.Key("warnings"): t.List(
                t.Dict(
                    {
                        t.Key("title"): t.String,
                        t.Key("message"): t.String,
                        t.Key("type"): t.String,
                    }
                ),
            ),
        }
    ).allow_extra("*")

    _converter = t.Dict(
        {
            t.Key("aggregation_type", optional=True): t.Null() | String(),
            t.Key("allowed_pairwise_interaction_groups_filename", optional=True): t.Null()
            | String(),
            t.Key("allow_partial_history_time_series_predictions", optional=True): t.Null()
            | t.Bool(),
            t.Key("autopilot_data_selection_method", optional=True): t.Null() | String(),
            t.Key("auto_start", optional=True): t.Null() | t.Bool(),
            t.Key("backtests", optional=True): t.Null() | t.List(_backtest_specification_converter),
            t.Key("calendar_id", optional=True): t.Null() | String(),
            t.Key("calendar_name", optional=True): t.Null() | String(),
            t.Key("class_mapping_aggregation_settings", optional=True): t.Null()
            | t.List(_class_mapping_aggregation_settings_converter),
            t.Key("class_mapping_aggregation_settings_enabled", optional=True): t.Null() | t.Bool(),
            t.Key("cross_series_group_by_columns", optional=True): t.Null() | t.List(String()),
            t.Key("cv_holdout_level", optional=True): t.Null() | t.Or(String(), Int()),
            t.Key("cv_method", optional=True): t.Null() | cv_method_converter,
            t.Key("date_removal", optional=True): t.Null() | t.Bool(),
            t.Key("datetime_partition_column", optional=True): t.Null() | String(),
            t.Key("datetime_partitioning_id", optional=True): t.Or(t.Null(), String()),
            t.Key("default_to_a_priori", optional=True): t.Null() | t.Bool(),
            t.Key("default_to_do_not_derive", optional=True): t.Null() | t.Bool(),
            t.Key("differencing_method", optional=True): t.Null() | String(),
            t.Key("disable_holdout", optional=True): t.Null() | t.Bool(),
            t.Key("error_message", optional=True): t.Null() | String(),
            t.Key("external_predictions", optional=True): t.Null() | t.List(Feature),
            t.Key("external_time_series_baseline_dataset_name", optional=True): t.Null() | String(),
            t.Key("feature_derivation_window_end", optional=True): t.Null() | Int(),
            t.Key("feature_derivation_window_start", optional=True): t.Null() | Int(),
            t.Key("feature_engineering_graphs", optional=True): t.Null()
            | t.List(RelationshipGraph._converter),
            t.Key("feature_engineering_options", optional=True): t.Or(
                t.Null(), t.Dict({}), FeatureEngineeringOptions._converter
            ),
            t.Key("feature_engineering_prediction_point", optional=True): t.Null()
            | Feature._converter,
            t.Key("featurelist_id", optional=True): t.Null() | String(),
            t.Key("feature_settings", optional=True): t.Null()
            | t.List(_feature_settings_converter),
            t.Key("forecast_window_end", optional=True): t.Null() | Int(),
            t.Key("forecast_window_start", optional=True): t.Null() | Int(),
            t.Key("gap_duration", optional=True): t.Null() | String(),
            t.Key("holdout_duration", optional=True): t.Or(t.Null(), String(), Duration._converter),
            t.Key("holdout_end_date", optional=True): t.Null() | parse_time,
            t.Key("holdout_level", optional=True): t.Null() | t.Or(String(), Int()),
            t.Key("holdout_pct", optional=True): t.Null() | Int(),
            t.Key("holdout_start_date", optional=True): t.Null() | parse_time,
            t.Key("initial_gpu_workers", optional=True): t.Null() | Int(),
            t.Key("initial_mode", optional=True): t.Null()
            | t.Or(
                t.Enum(AUTOPILOT_MODE.FULL_AUTO),
                t.Enum(AUTOPILOT_MODE.MANUAL),
                t.Enum(AUTOPILOT_MODE.QUICK),
                t.Enum(AUTOPILOT_MODE.COMPREHENSIVE),
            ),
            t.Key("initial_num_workers", optional=True): t.Null() | Int(),
            t.Key("is_dirty", optional=True): t.Null() | t.Bool(),
            t.Key("is_holdout_modified", optional=True): t.Null() | t.Bool(),
            t.Key("metric", optional=True): t.Null() | String(),
            t.Key("model_splits", optional=True): t.Null() | Int(),
            t.Key("multiseries_id_columns", optional=True): t.Null() | t.List(String()),
            t.Key("number_of_backtests", optional=True): t.Null() | Int(),
            t.Key("partitioning_warnings", optional=True): t.Or(
                t.Null() | t.List(_partitioning_warnings_converter)
            ),
            t.Key("partitioning_extended_warnings", optional=True): t.Or(
                t.Null() | t.List(_partitioning_extended_warnings_converter)
            ),
            t.Key("partition_key_cols", optional=True): t.Null() | String(),
            t.Key("periodicities", optional=True): t.Null() | t.List(Periodicity._converter),
            t.Key("positive_class", optional=True): t.Null() | t.Or(String(), Int()),
            t.Key("project_id", optional=True): t.Null() | String(),
            t.Key("quintile_level", optional=True): t.Null() | Int(),
            t.Key("relationships_configuration_id", optional=True): t.Null() | String(),
            t.Key("reps", optional=True): t.Null() | Int(),
            t.Key("sample_step_pct", optional=True): t.Null() | Int(),
            t.Key("segmentation_id_column", optional=True): t.Null() | String(),
            t.Key("segmentation_model_package_id", optional=True): t.Null() | String(),
            t.Key("segmentation_model_package_name", optional=True): t.Null() | String(),
            t.Key("segmentation_task_id", optional=True): t.Null() | String(),
            t.Key("segments_count", optional=True): t.Null() | Int(),
            t.Key("target", optional=True): t.Null() | String(),
            t.Key("target_type", optional=True): t.Null() | t.Enum(*TARGET_TYPE.ALL),
            t.Key("training_level", optional=True): t.Null() | t.Or(String(), Int()),
            t.Key("treat_as_exponential", optional=True): t.Null()
            | t.Enum(*TREAT_AS_EXPONENTIAL.ALL),
            t.Key("use_gpu", optional=True): t.Null() | t.Bool(),
            t.Key("unsupervised_mode", optional=True): t.Null() | t.Bool(),
            t.Key("use_cross_series_features", optional=True): t.Null() | t.Bool(),
            t.Key("use_project_settings", optional=True): t.Null() | t.Bool(),
            t.Key("user_partition_col", optional=True): t.Null() | Feature._converter,
            t.Key("use_time_series", optional=True): t.Null() | t.Bool(),
            t.Key("validation_duration", optional=True): t.Null() | String(),
            t.Key("validation_level", optional=True): t.Null() | t.Or(String(), Int()),
            t.Key("validation_pct", optional=True): t.Null() | Int(),
            t.Key("validation_type", optional=True): t.Null() | t.Enum(*VALIDATION_TYPE.ALL),
            t.Key("windows_basis_unit", optional=True): t.Null() | t.Enum(*TIME_UNITS.ALL),
        }
    ).allow_extra("*")

    # The super() call is included in the _set_values method
    def __init__(  # pylint: disable=W0231
        self,
        aggregation_type: Optional[str] = None,
        allowed_pairwise_interaction_groups_filename: Optional[str] = None,
        allow_partial_history_time_series_predictions: Optional[bool] = None,
        autopilot_data_selection_method: Optional[str] = None,
        auto_start: Optional[bool] = None,
        backtests: Optional[List[BacktestSpecification]] = None,
        calendar_id: Optional[str] = None,
        calendar_name: Optional[str] = None,
        class_mapping_aggregation_settings: Optional[List[ClassMappingAggregationSettings]] = None,
        class_mapping_aggregation_settings_enabled: Optional[bool] = None,
        cross_series_group_by_columns: Optional[List[str]] = None,
        cv_holdout_level: Optional[Union[str, int]] = None,
        cv_method: Optional[CV_METHOD] = None,
        date_removal: Optional[bool] = None,
        datetime_partition_column: Optional[str] = None,
        datetime_partitioning_id: Optional[str] = None,
        default_to_a_priori: Optional[bool] = None,
        default_to_do_not_derive: Optional[bool] = None,
        differencing_method: Optional[str] = None,
        disable_holdout: Optional[bool] = None,
        error_message: Optional[str] = None,
        external_predictions: Optional[List[Feature]] = None,
        external_time_series_baseline_dataset_name: Optional[str] = None,
        feature_derivation_window_end: Optional[int] = None,
        feature_derivation_window_start: Optional[int] = None,
        feature_engineering_graphs: Optional[List[RelationshipGraph]] = None,
        feature_engineering_options: Optional[FeatureEngineeringOptions] = None,
        feature_engineering_prediction_point: Optional[Feature] = None,
        featurelist_id: Optional[str] = None,
        feature_settings: Optional[List[FeatureSettings]] = None,
        forecast_window_end: Optional[int] = None,
        forecast_window_start: Optional[int] = None,
        gap_duration: Optional[str] = None,
        holdout_duration: Optional[Duration] = None,
        holdout_end_date: Optional[str] = None,
        holdout_level: Optional[Union[str, int]] = None,
        holdout_pct: Optional[int] = None,
        holdout_start_date: Optional[str] = None,
        initial_gpu_workers: Optional[int] = None,
        initial_mode: Optional[
            Union[
                AUTOPILOT_MODE.QUICK,
                AUTOPILOT_MODE.FULL_AUTO,
                AUTOPILOT_MODE.MANUAL,
                AUTOPILOT_MODE.COMPREHENSIVE,
            ]
        ] = None,
        initial_num_workers: Optional[int] = None,
        is_dirty: Optional[bool] = None,
        is_holdout_modified: Optional[bool] = None,
        metric: Optional[str] = None,
        model_splits: Optional[int] = None,
        multiseries_id_columns: Optional[List[str]] = None,
        number_of_backtests: Optional[int] = None,
        partitioning_warnings: Optional[List[PartitioningWarnings]] = None,
        partitioning_extended_warnings: Optional[List[PartitioningExtendedWarnings]] = None,
        partition_key_cols: Optional[str] = None,
        periodicities: Optional[List[Periodicity]] = None,
        positive_class: Optional[Union[str, int]] = None,
        project_id: Optional[str] = None,
        quintile_level: Optional[int] = None,
        relationships_configuration_id: Optional[str] = None,
        reps: Optional[int] = None,
        sample_step_pct: Optional[int] = None,
        segmentation_id_column: Optional[str] = None,
        segmentation_model_package_id: Optional[str] = None,
        segmentation_model_package_name: Optional[str] = None,
        segmentation_task_id: Optional[str] = None,
        segments_count: Optional[int] = None,
        target: Optional[str] = None,
        target_type: Optional[TARGET_TYPE] = None,
        training_level: Optional[Union[str, int]] = None,
        treat_as_exponential: Optional[TREAT_AS_EXPONENTIAL] = None,
        use_gpu: Optional[bool] = None,
        unsupervised_mode: Optional[bool] = None,
        use_cross_series_features: Optional[bool] = None,
        use_project_settings: Optional[bool] = None,
        user_partition_col: Optional[Feature] = None,
        use_time_series: Optional[bool] = None,
        validation_duration: Optional[str] = None,
        validation_level: Optional[Union[str, int]] = None,
        validation_pct: Optional[int] = None,
        validation_type: Optional[VALIDATION_TYPE] = None,
        windows_basis_unit: Optional[TIME_UNITS] = None,
        **kwargs,
    ):
        self._set_values(
            aggregation_type=aggregation_type,
            allowed_pairwise_interaction_groups_filename=allowed_pairwise_interaction_groups_filename,
            allow_partial_history_time_series_predictions=allow_partial_history_time_series_predictions,
            autopilot_data_selection_method=autopilot_data_selection_method,
            auto_start=auto_start,
            backtests=backtests,
            calendar_id=calendar_id,
            calendar_name=calendar_name,
            class_mapping_aggregation_settings=class_mapping_aggregation_settings,
            class_mapping_aggregation_settings_enabled=class_mapping_aggregation_settings_enabled,
            cross_series_group_by_columns=cross_series_group_by_columns,
            cv_holdout_level=cv_holdout_level,
            cv_method=cv_method,
            date_removal=date_removal,
            datetime_partition_column=datetime_partition_column,
            datetime_partitioning_id=datetime_partitioning_id,
            default_to_a_priori=default_to_a_priori,
            default_to_do_not_derive=default_to_do_not_derive,
            differencing_method=differencing_method,
            disable_holdout=disable_holdout,
            error_message=error_message,
            external_predictions=external_predictions,
            external_time_series_baseline_dataset_name=external_time_series_baseline_dataset_name,
            feature_derivation_window_end=feature_derivation_window_end,
            feature_derivation_window_start=feature_derivation_window_start,
            feature_engineering_graphs=feature_engineering_graphs,
            feature_engineering_options=feature_engineering_options,
            feature_engineering_prediction_point=feature_engineering_prediction_point,
            featurelist_id=featurelist_id,
            feature_settings=feature_settings,
            forecast_window_end=forecast_window_end,
            forecast_window_start=forecast_window_start,
            gap_duration=gap_duration,
            holdout_duration=holdout_duration,
            holdout_end_date=holdout_end_date,
            holdout_level=holdout_level,
            holdout_pct=holdout_pct,
            holdout_start_date=holdout_start_date,
            initial_gpu_workers=initial_gpu_workers,
            initial_mode=initial_mode,
            initial_num_workers=initial_num_workers,
            is_dirty=is_dirty,
            is_holdout_modified=is_holdout_modified,
            metric=metric,
            model_splits=model_splits,
            multiseries_id_columns=multiseries_id_columns,
            number_of_backtests=number_of_backtests,
            partitioning_warnings=partitioning_warnings,
            partitioning_extended_warnings=partitioning_extended_warnings,
            partition_key_cols=partition_key_cols,
            periodicities=periodicities,
            positive_class=positive_class,
            project_id=project_id,
            quintile_level=quintile_level,
            relationships_configuration_id=relationships_configuration_id,
            reps=reps,
            sample_step_pct=sample_step_pct,
            segmentation_id_column=segmentation_id_column,
            segmentation_model_package_id=segmentation_model_package_id,
            segmentation_model_package_name=segmentation_model_package_name,
            segmentation_task_id=segmentation_task_id,
            segments_count=segments_count,
            target=target,
            target_type=target_type,
            training_level=training_level,
            treat_as_exponential=treat_as_exponential,
            use_gpu=use_gpu,
            unsupervised_mode=unsupervised_mode,
            use_cross_series_features=use_cross_series_features,
            use_project_settings=use_project_settings,
            user_partition_col=user_partition_col,
            use_time_series=use_time_series,
            validation_duration=validation_duration,
            validation_level=validation_level,
            validation_pct=validation_pct,
            validation_type=validation_type,
            windows_basis_unit=windows_basis_unit,
            **kwargs,
        )

    def _set_values(
        self,
        aggregation_type: Optional[str] = None,
        allowed_pairwise_interaction_groups_filename: Optional[str] = None,
        allow_partial_history_time_series_predictions: Optional[bool] = None,
        autopilot_data_selection_method: Optional[str] = None,
        auto_start: Optional[bool] = None,
        backtests: Optional[List[BacktestSpecification]] = None,
        calendar_id: Optional[str] = None,
        calendar_name: Optional[str] = None,
        class_mapping_aggregation_settings: Optional[List[ClassMappingAggregationSettings]] = None,
        class_mapping_aggregation_settings_enabled: Optional[bool] = None,
        cross_series_group_by_columns: Optional[List[str]] = None,
        cv_holdout_level: Optional[Union[str, int]] = None,
        cv_method: Optional[str] = None,
        date_removal: Optional[bool] = None,
        datetime_partition_column: Optional[str] = None,
        datetime_partitioning_id: Optional[str] = None,
        default_to_a_priori: Optional[bool] = None,
        default_to_do_not_derive: Optional[bool] = None,
        differencing_method: Optional[str] = None,
        disable_holdout: Optional[bool] = None,
        error_message: Optional[str] = None,
        external_predictions: Optional[List[Feature]] = None,
        external_time_series_baseline_dataset_name: Optional[str] = None,
        feature_derivation_window_end: Optional[int] = None,
        feature_derivation_window_start: Optional[int] = None,
        feature_engineering_graphs: Optional[List[RelationshipGraph]] = None,
        feature_engineering_options: Optional[FeatureEngineeringOptions] = None,
        feature_engineering_prediction_point: Optional[Feature] = None,
        featurelist_id: Optional[str] = None,
        feature_settings: Optional[List[FeatureSettings]] = None,
        forecast_window_end: Optional[int] = None,
        forecast_window_start: Optional[int] = None,
        gap_duration: Optional[str] = None,
        holdout_duration: Optional[Duration] = None,
        holdout_end_date: Optional[str] = None,
        holdout_level: Optional[Union[str, int]] = None,
        holdout_pct: Optional[int] = None,
        holdout_start_date: Optional[str] = None,
        initial_gpu_workers: Optional[int] = None,
        initial_mode: Optional[
            Union[AUTOPILOT_MODE.FULL_AUTO, AUTOPILOT_MODE.MANUAL, AUTOPILOT_MODE.COMPREHENSIVE]
        ] = None,
        initial_num_workers: Optional[int] = None,
        is_dirty: Optional[bool] = None,
        is_holdout_modified: Optional[bool] = None,
        metric: Optional[str] = None,
        model_splits: Optional[int] = None,
        multiseries_id_columns: Optional[List[str]] = None,
        number_of_backtests: Optional[int] = None,
        partitioning_warnings: Optional[List[PartitioningWarnings]] = None,
        partitioning_extended_warnings: Optional[List[PartitioningExtendedWarnings]] = None,
        partition_key_cols: Optional[str] = None,
        periodicities: Optional[List[Periodicity]] = None,
        positive_class: Optional[Union[str, int]] = None,
        project_id: Optional[str] = None,
        quintile_level: Optional[int] = None,
        relationships_configuration_id: Optional[str] = None,
        reps: Optional[int] = None,
        sample_step_pct: Optional[int] = None,
        segmentation_id_column: Optional[str] = None,
        segmentation_model_package_id: Optional[str] = None,
        segmentation_model_package_name: Optional[str] = None,
        segmentation_task_id: Optional[str] = None,
        segments_count: Optional[int] = None,
        target: Optional[str] = None,
        target_type: Optional[TARGET_TYPE] = None,
        training_level: Optional[Union[str, int]] = None,
        treat_as_exponential: Optional[TREAT_AS_EXPONENTIAL] = None,
        use_gpu: Optional[bool] = None,
        unsupervised_mode: Optional[bool] = None,
        use_cross_series_features: Optional[bool] = None,
        use_project_settings: Optional[bool] = None,
        user_partition_col: Optional[Feature] = None,
        use_time_series: Optional[bool] = None,
        validation_duration: Optional[str] = None,
        validation_level: Optional[Union[str, int]] = None,
        validation_pct: Optional[int] = None,
        validation_type: Optional[VALIDATION_TYPE] = None,
        windows_basis_unit: Optional[TIME_UNITS] = None,
        **kwargs,
    ):
        """
        A helper method to set all instance values. See class definition for value types and descriptions.
        """
        self.aggregation_type = aggregation_type
        self.allowed_pairwise_interaction_groups_filename = (
            allowed_pairwise_interaction_groups_filename
        )
        self.allow_partial_history_time_series_predictions = (
            allow_partial_history_time_series_predictions
        )
        self.autopilot_data_selection_method = autopilot_data_selection_method
        self.auto_start = auto_start
        if backtests and isinstance(backtests[0], dict):
            self.backtests = [BacktestSpecification(**backtest) for backtest in backtests]
        else:
            self.backtests = backtests
        self.calendar_id = calendar_id
        self.calendar_name = calendar_name
        if class_mapping_aggregation_settings and isinstance(
            class_mapping_aggregation_settings[0], dict
        ):
            self.class_mapping_aggregation_settings = [
                ClassMappingAggregationSettings(**setting)
                for setting in class_mapping_aggregation_settings
            ]
        else:
            self.class_mapping_aggregation_settings = class_mapping_aggregation_settings
        self.class_mapping_aggregation_settings_enabled = class_mapping_aggregation_settings_enabled
        self.cross_series_group_by_columns = cross_series_group_by_columns
        self.cv_holdout_level = cv_holdout_level
        self.cv_method = cv_method
        self.date_removal = date_removal
        self.datetime_partition_column = datetime_partition_column
        self.datetime_partitioning_id = datetime_partitioning_id
        self.default_to_a_priori = default_to_a_priori
        self.default_to_do_not_derive = default_to_do_not_derive
        self.differencing_method = differencing_method
        self.disable_holdout = disable_holdout
        self.error_message = error_message
        if external_predictions and isinstance(external_predictions[0], dict):
            self.external_predictions = [
                Feature(**external_prediction) for external_prediction in external_predictions
            ]
        else:
            self.external_predictions = external_predictions
        self.external_time_series_baseline_dataset_name = external_time_series_baseline_dataset_name
        self.feature_derivation_window_end = feature_derivation_window_end
        self.feature_derivation_window_start = feature_derivation_window_start
        if feature_engineering_graphs and isinstance(feature_engineering_graphs[0], dict):
            self.feature_engineering_graphs = [
                RelationshipGraph(**graph) for graph in feature_engineering_graphs
            ]
        else:
            self.feature_engineering_graphs = feature_engineering_graphs
        if feature_engineering_options and isinstance(feature_engineering_options, dict):
            self.feature_engineering_options = FeatureEngineeringOptions(
                **feature_engineering_options
            )
        else:
            self.feature_engineering_options = feature_engineering_options
        self.feature_engineering_prediction_point = feature_engineering_prediction_point
        self.featurelist_id = featurelist_id
        if feature_settings and isinstance(feature_settings[0], dict):
            self.feature_settings = [
                FeatureSettings(
                    **feature_setting
                    if "a_priori" not in feature_setting.keys()
                    else {"known_in_advance": feature_setting.pop("a_priori"), **feature_setting}
                )
                for feature_setting in feature_settings
            ]
        else:
            self.feature_settings = feature_settings
        self.forecast_window_end = forecast_window_end
        self.forecast_window_start = forecast_window_start
        self.gap_duration = gap_duration
        if holdout_duration and isinstance(holdout_duration, dict):
            self.holdout_duration = Duration(**holdout_duration)
        else:
            self.holdout_duration = holdout_duration
        self.holdout_end_date = holdout_end_date
        self.holdout_level = holdout_level
        self.holdout_pct = holdout_pct
        self.holdout_start_date = holdout_start_date
        self.initial_gpu_workers = initial_gpu_workers
        self.initial_mode = initial_mode
        self.initial_num_workers = initial_num_workers
        self.is_dirty = is_dirty
        self.is_holdout_modified = is_holdout_modified
        self.metric = metric
        self.model_splits = model_splits
        self.multiseries_id_columns = multiseries_id_columns
        self.number_of_backtests = number_of_backtests
        self.partitioning_warnings = partitioning_warnings
        self.partitioning_extended_warnings = partitioning_extended_warnings
        self.partition_key_cols = partition_key_cols
        if periodicities and isinstance(periodicities[0], dict):
            self.periodicities = [Periodicity(**periodicity) for periodicity in periodicities]
        else:
            self.periodicities = periodicities
        self.positive_class = positive_class
        self.project_id = project_id
        self.quintile_level = quintile_level
        self.relationships_configuration_id = relationships_configuration_id
        self.reps = reps
        self.sample_step_pct = sample_step_pct
        self.segmentation_id_column = segmentation_id_column
        self.segmentation_model_package_id = segmentation_model_package_id
        self.segmentation_model_package_name = segmentation_model_package_name
        self.segmentation_task_id = segmentation_task_id
        self.segments_count = segments_count
        self.target = target
        self.target_type = target_type
        self.training_level = training_level
        self.treat_as_exponential = treat_as_exponential
        self.use_gpu = use_gpu
        self.unsupervised_mode = unsupervised_mode
        self.use_cross_series_features = use_cross_series_features
        self.use_project_settings = use_project_settings
        if user_partition_col and isinstance(user_partition_col, dict):
            self.user_partition_col = Feature(**user_partition_col)
        else:
            self.user_partition_col = user_partition_col
        self.use_time_series = use_time_series
        self.validation_duration = validation_duration
        self.validation_level = validation_level
        self.validation_pct = validation_pct
        self.validation_type = validation_type
        self.windows_basis_unit = windows_basis_unit
        super().__init__(**kwargs)

    @property
    def is_empty(self) -> bool:
        return all(
            (value is None or key in self._fields_with_defaults)
            for key, value in vars(self).items()
        )

    @classmethod
    def get(cls, project_id: str) -> ProjectOptions:
        """
        Return a ProjectOptions instance containing the stored options for the project with the ID ``project_id``.

        Parameters
        ----------
        project_id : str : The ID of the project that holds the stored options and settings you are trying to access.

        Returns
        -------
        A ProjectOptions instance.
        """
        try:
            path = f"projects/{project_id}/options/"
            new_project_options = cls.from_location(path)
            if new_project_options.project_id is None:
                # Newly created projects return an empty json object, so we need to seed this
                # new project options instance with the correct project_id
                new_project_options = cls(project_id=project_id)
            return new_project_options
        except ClientError as ex:
            # If we receive a 404, it's because this project doesn't exist in DataRobot yet.
            if ex.status_code == 404:
                raise ClientError(
                    "Project options for this project do not exist. "
                    "Please create a project in DataRobot first in order to retrieve the project's options.",
                    status_code=ex.status_code,
                )
            # If it's something other than a 404, the error needs to be raised to the
            # user.
            else:
                raise ex

    def collect_payload(self) -> Dict[str, Any]:
        return {
            "accuracy_optimized_mb": self.accuracy_optimized_mb,
            "autopilot_with_feature_discovery": self.autopilot_with_feature_discovery,
            "auto_start": self.auto_start,
            "backtests": [backtest.collect_payload() for backtest in self.backtests]
            if self.backtests
            else None,
            "blend_best_models": self.blend_best_models,
            "blueprint_threshold": self.blueprint_threshold,
            "class_mapping_aggregation_settings": [
                setting.collect_payload() for setting in self.class_mapping_aggregation_settings
            ]
            if self.class_mapping_aggregation_settings
            else None,
            "class_mapping_aggregation_settings_enabled": self.class_mapping_aggregation_settings_enabled,
            "cv_method": self.cv_method,
            "datetime_partition_column": self.datetime_partition_column,
            "datetime_partitioning_id": self.datetime_partitioning_id,
            "differencing_method": self.differencing_method,
            "disable_holdout": self.disable_holdout,
            "feature_derivation_window_end": self.feature_derivation_window_end,
            "feature_derivation_window_start": self.feature_derivation_window_start,
            "feature_discovery_supervised_feature_reduction": self.feature_discovery_supervised_feature_reduction,
            "feature_engineering_options": self.feature_engineering_options.collect_payload()
            if self.feature_engineering_options
            else None,
            "featurelist_id": self.featurelist_id,
            "feature_settings": [
                setting.collect_payload(use_a_priori=True) for setting in self.feature_settings
            ]
            if self.feature_settings
            else None,
            "forecast_window_end": self.forecast_window_end,
            "forecast_window_start": self.forecast_window_start,
            "gap_duration": self.gap_duration,
            "holdout_duration": self.holdout_duration,
            "holdout_end_date": self.holdout_end_date,
            "holdout_start_date": self.holdout_start_date,
            "initial_mode": self.initial_mode,
            "is_dirty": self.is_dirty,
            "is_holdout_modified": self.is_holdout_modified,
            "metric": self.metric,
            "model_splits": self.model_splits,
            "number_of_backtests": self.number_of_backtests,
            "prepare_model_for_deployment": self.prepare_model_for_deployment,
            "run_leakage_removed_feature_list": self.run_leakage_removed_feature_list,
            "smart_downsampled": self.smart_downsampled,
            "target": self.target,
            "treat_as_exponential": self.treat_as_exponential,
            "use_gpu": self.use_gpu,
            "unsupervised_mode": self.unsupervised_mode,
            "use_supervised_feature_reduction": self.use_supervised_feature_reduction,
            "use_time_series": self.use_time_series,
            "validation_duration": self.validation_duration,
            "windows_basis_unit": self.windows_basis_unit,
        }

    def collect_autopilot_payload(self) -> Dict[str, Any]:
        if not self.is_empty:
            return super().collect_payload()
        return {
            default_field: getattr(self, default_field)
            for default_field in self._fields_with_defaults
        }

    def update_options(self, use_patch=True) -> Optional[ProjectOptions]:
        """
        Parameters
        ----------
        use_patch : bool : optional
            Switch between using a PATCH and PUT call to update project options.
            PATCH will return the updated object instance, while PUT will return None.
        Returns
        -------
        Updated ProjectOptions class instance or None
        """
        url = f"projects/{self.project_id}/options/"
        data = self.collect_payload()
        if not use_patch:
            self._client.put(url, data=data)
            return None
        response = self._client.patch(url, data=data)
        data = response.json()
        self._set_values(**from_api(data, do_recursive=True))
        return self



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
"""
Trafaret validators that doesn't part of the class.
Separated to own file to improve readability and reusability.
"""

import trafaret as t

from datarobot._compat import Int, String

# Representation of a single feature impact
# featureName - feature we are describing
# impactUnnormalized - drop in the metric value comparing to base
# impactNormalized - normalized value so that highest drop = 1.0
# redundantWith - one of 2 options
# - None if redundant detection wasn't run or
# - feature name that makes this one redundant
single_feature_impact_trafaret = t.Dict(
    {
        t.Key("impactUnnormalized"): t.Float,
        t.Key("impactNormalized"): t.Float,
        t.Key("redundantWith"): t.Or(t.Null, String),
        t.Key("featureName"): String,
    }
).ignore_extra("*")

# Feature impact data for regression, binary classification and aggregated score for multiclass
# ranRedundancyDetection - shows if redundancy detection was run for the model
# featureImpacts - list of impacts for each feature
feature_impact_trafaret = t.Dict(
    {
        t.Key("count"): Int,
        t.Key("ranRedundancyDetection"): t.Bool,
        t.Key("rowCount"): t.Or(Int, t.Null),
        t.Key("shapBased"): t.Bool,
        # to make newer client work with older DataRobot responses
        t.Key("backtest", optional=True, default=None): t.Or(Int(gte=0), t.Null, t.Atom("holdout")),
        t.Key("featureImpacts"): t.List(single_feature_impact_trafaret),
        t.Key("dataSliceId", optional=True, default=None): t.Or(String, t.Null),
    }
).ignore_extra("*")


custom_model_feature_impact_trafaret = t.Dict(
    {
        t.Key("count"): Int,
        t.Key("ranRedundancyDetection"): t.Bool,
        t.Key("rowCount"): t.Or(Int, t.Null),
        t.Key("shapBased"): t.Bool,
        t.Key("featureImpacts"): t.List(single_feature_impact_trafaret),
    }
).ignore_extra("*")


# Feature impact data for each class in multiclass model
# classFeatureImpacts - list of records for each class, with 2 keys
# - class - target class name
# - featureImpacts - list of impacts of each feature for the class
multiclass_feature_impact_trafaret = t.Dict(
    {
        t.Key("ranRedundancyDetection"): t.Bool,
        t.Key("classFeatureImpacts"): t.List(
            t.Dict(
                {
                    t.Key("featureImpacts"): t.List(single_feature_impact_trafaret),
                    t.Key("class"): String,
                }
            ).ignore_extra("*")
        ),
    }
).ignore_extra("*")



#
# Copyright 2021-2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

import contextlib
import copy
import json
import os
import time
from typing import Any, cast, Dict, Iterable, List, Mapping, Optional, Tuple, Union

from requests import Response
from requests_toolbelt import MultipartEncoder
import trafaret as t

from datarobot._compat import Int, String
from datarobot.enums import DEFAULT_MAX_WAIT, NETWORK_EGRESS_POLICY
from datarobot.errors import InvalidUsageError, TrainingDataAssignmentError
from datarobot.models.api_object import APIObject, ServerDataType
from datarobot.models.job import filter_feature_impact_result
from datarobot.models.validators import custom_model_feature_impact_trafaret
from datarobot.utils import camelize
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution, wait_for_custom_resolution


class RequiredMetadataValue(APIObject):
    """Definition of a metadata key that custom models using this environment must define

    .. versionadded:: v2.26

    Attributes
    ----------
    field_name: str
        The required field names.  Required field names are defined by the
        environment's required_metadata_keys. This value will be added as an
        environment vairable when running custom models.
    value: str
        The value for the required field.
    """

    _converter = t.Dict({t.Key("field_name"): String(), t.Key("value"): String()})

    schema = _converter

    def __init__(self, **kwargs: Any) -> None:
        self._set_values(**kwargs)

    def __repr__(self) -> str:
        return "{}(field_name={!r}, value={!r})".format(
            self.__class__.__name__,
            self.field_name,
            self.value,
        )

    def _set_values(self, field_name: str, value: str) -> None:
        self.field_name = field_name
        self.value = value

    def to_dict(self) -> Dict[str, str]:
        return cast(
            Dict[str, str],
            self._converter.check({"field_name": self.field_name, "value": self.value}),
        )


class TrainingData(APIObject):
    """Training data assigned to a DataRobot custom model version.

    .. versionadded:: v3.2

    Attributes
    ----------
    dataset_id: str
        The ID of the dataset.
    dataset_version_id: str
        The ID of the dataset version.
    dataset_name: str
        The name of the dataset.
    assignment_in_progress: bool
        The status of the assignment in progress.
    assignment_error: dict
        The assignment error message.
    """

    _converter = t.Dict(
        {
            t.Key("dataset_id", optional=True): t.Or(String(), t.Null()),
            t.Key("dataset_version_id", optional=True): t.Or(String(), t.Null()),
            t.Key("dataset_name", optional=True): t.Or(String(), t.Null()),
            t.Key("assignment_in_progress", optional=True): t.Bool(),
            t.Key("assignment_error", optional=True): t.Dict(
                {
                    t.Key("message"): t.Or(String(), t.Null()),
                }
            )
            | t.Null,
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(
        self,
        dataset_id: Optional[str] = None,
        dataset_version_id: Optional[str] = None,
        dataset_name: Optional[str] = None,
        assignment_in_progress: Optional[str] = None,
        assignment_error: Optional[Dict[str, str]] = None,
    ) -> None:
        self.dataset_id = dataset_id
        self.dataset_version_id = dataset_version_id
        self.dataset_name = dataset_name
        self.assignment_in_progress = assignment_in_progress
        self.assignment_error = assignment_error


class HoldoutData(APIObject):
    """Holdout data assigned to a DataRobot custom model version.

    .. versionadded:: v3.2

    Attributes
    ----------
    dataset_id: str
        The ID of the dataset.
    dataset_version_id: str
        The ID of the dataset version.
    dataset_name: str
        The name of the dataset.
    partition_column: str
        The name of the partitions column.
    """

    _converter = t.Dict(
        {
            t.Key("dataset_id", optional=True): t.Or(String(), t.Null()),
            t.Key("dataset_version_id", optional=True): t.Or(String(), t.Null()),
            t.Key("dataset_name", optional=True): t.Or(String(), t.Null()),
            t.Key("partition_column", optional=True): t.Or(String(), t.Null()),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(
        self,
        dataset_id: Optional[str] = None,
        dataset_version_id: Optional[str] = None,
        dataset_name: Optional[str] = None,
        partition_column: Optional[str] = None,
    ) -> None:
        self.dataset_id = dataset_id
        self.dataset_version_id = dataset_version_id
        self.dataset_name = dataset_name
        self.partition_column = partition_column


class CustomModelFileItem(APIObject):
    """A file item attached to a DataRobot custom model version.

    .. versionadded:: v2.21

    Attributes
    ----------
    id: str
        The ID of the file item.
    file_name: str
        The name of the file item.
    file_path: str
        The path of the file item.
    file_source: str
        The source of the file item.
    created_at: str, optional
        ISO-8601 formatted timestamp of when the version was created.
    """

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("file_name"): String(),
            t.Key("file_path"): String(),
            t.Key("file_source"): String(),
            t.Key("created", optional=True) >> "created_at": String(),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(
        self,
        id: str,
        file_name: str,
        file_path: str,
        file_source: str,
        created_at: Optional[str] = None,
    ) -> None:
        self.id = id
        self.file_name = file_name
        self.file_path = file_path
        self.file_source = file_source
        self.created_at = created_at


class CustomModelVersionDependencyBuild(APIObject):
    """Metadata about a DataRobot custom model version's dependency build

    .. versionadded:: v2.22

    Attributes
    ----------
    custom_model_id: str
        The ID of the custom model.
    custom_model_version_id: str
        The ID of the custom model version.
    build_status: str
        The status of the custom model version's dependency build.
    started_at: str
        ISO-8601 formatted timestamp of when the build was started.
    completed_at: str, optional
        ISO-8601 formatted timestamp of when the build has completed.
    """

    _path = "customModels/{}/versions/{}/dependencyBuild/"
    _log_path = "customModels/{}/versions/{}/dependencyBuildLog/"

    _converter = t.Dict(
        {
            t.Key("custom_model_id"): String(),
            t.Key("custom_model_version_id"): String(),
            t.Key("build_status"): String(),
            t.Key("build_start") >> "started_at": String(),
            t.Key("build_end", optional=True) >> "completed_at": String(allow_blank=True),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self, **kwargs: Any) -> None:
        self._set_values(**kwargs)

    def __repr__(self) -> str:
        return "{}(model={!r}, version={!r}, status={!r})".format(
            self.__class__.__name__,
            self.custom_model_id,
            self.custom_model_version_id,
            self.build_status,
        )

    def _set_values(  # pylint: disable=missing-function-docstring
        self,
        custom_model_id: str,
        custom_model_version_id: str,
        build_status: str,
        started_at: str,
        completed_at: Optional[str] = None,
    ) -> None:
        self.custom_model_id = custom_model_id
        self.custom_model_version_id = custom_model_version_id
        self.build_status = build_status
        self.started_at = started_at
        self.completed_at = completed_at

    @classmethod
    def _update_server_data(
        cls, server_data: Dict[str, Any], custom_model_id: str, custom_model_version_id: str
    ) -> Dict[str, Any]:
        updated_data = copy.copy(server_data)
        updated_data.update(
            {"customModelId": custom_model_id, "customModelVersionId": custom_model_version_id}
        )
        return updated_data

    @classmethod
    def get_build_info(
        cls, custom_model_id: str, custom_model_version_id: str
    ) -> CustomModelVersionDependencyBuild:
        """Retrieve information about a custom model version's dependency build

        .. versionadded:: v2.22

        Parameters
        ----------
        custom_model_id: str
            The ID of the custom model.
        custom_model_version_id: str
            The ID of the custom model version.

        Returns
        -------
        CustomModelVersionDependencyBuild
            The dependency build information.
        """
        url = cls._path.format(custom_model_id, custom_model_version_id)
        response = cls._client.get(url)
        server_data = response.json()
        updated_data = cls._update_server_data(
            server_data, custom_model_id, custom_model_version_id
        )
        return cls.from_server_data(updated_data)

    @classmethod
    def start_build(
        cls,
        custom_model_id: str,
        custom_model_version_id: str,
        max_wait: Optional[int] = DEFAULT_MAX_WAIT,
    ) -> Optional[CustomModelVersionDependencyBuild]:
        """Start the dependency build for a custom model version  dependency build

        .. versionadded:: v2.22

        Parameters
        ----------
        custom_model_id: str
            The ID of the custom model
        custom_model_version_id: str
            the ID of the custom model version
        max_wait: int, optional
            Max time to wait for a build completion.
            If set to None - method will return without waiting.
        """

        def build_complete(response: Response) -> Optional[CustomModelVersionDependencyBuild]:
            data = response.json()
            if data["buildStatus"] in ["success", "failed"]:
                updated_data = cls._update_server_data(
                    data, custom_model_id, custom_model_version_id
                )
                return cls.from_server_data(updated_data)
            return None

        url = cls._path.format(custom_model_id, custom_model_version_id)
        response = cls._client.post(url)

        if max_wait is None:
            server_data = response.json()
            updated_data = cls._update_server_data(
                server_data, custom_model_id, custom_model_version_id
            )
            return cls.from_server_data(updated_data)
        else:
            return cast(
                Optional[CustomModelVersionDependencyBuild],
                wait_for_custom_resolution(cls._client, url, build_complete, max_wait),
            )

    def get_log(self) -> str:
        """Get log of a custom model version dependency build.

        .. versionadded:: v2.22

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        url = self._log_path.format(self.custom_model_id, self.custom_model_version_id)
        return cast(str, self._client.get(url).text)

    def cancel(self) -> None:
        """Cancel custom model version dependency build that is in progress.

        .. versionadded:: v2.22

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        url = self._path.format(self.custom_model_id, self.custom_model_version_id)
        self._client.delete(url)

    def refresh(self) -> None:
        """Update custom model version dependency build with the latest data from server.

        .. versionadded:: v2.22

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        url = self._path.format(self.custom_model_id, self.custom_model_version_id)

        response = self._client.get(url)

        data = response.json()
        updated_data = self._update_server_data(
            data, self.custom_model_id, self.custom_model_version_id
        )
        self._set_values(**self._safe_data(updated_data, do_recursive=True))  # type: ignore[no-untyped-call]


class CustomDependencyConstraint(APIObject):
    """Metadata about a constraint on a dependency of a custom model version

    .. versionadded:: v2.22

    Attributes
    ----------
    constraint_type: str
        How the dependency should be constrained by version (<, <=, ==, >=, >).
    version: str
        The version to use in the dependency's constraint.
    """

    _converter = t.Dict(
        {t.Key("constraint_type"): String(), t.Key("version"): String()}
    ).ignore_extra("*")

    schema = _converter

    def __init__(self, **kwargs: Any) -> None:
        self._set_values(**kwargs)

    def __repr__(self) -> str:
        return "{}(constraint_type={!r}, version={!r})".format(
            self.__class__.__name__,
            self.constraint_type,
            self.version,
        )

    def _set_values(self, constraint_type: str, version: str) -> None:
        self.constraint_type = constraint_type
        self.version = version


class CustomDependency(APIObject):
    """Metadata about an individual dependency of a custom model version

    .. versionadded:: v2.22

    Attributes
    ----------
    package_name: str
        The dependency's package name.
    constraints: List[CustomDependencyConstraint]
        Version constraints to apply on the dependency.
    line: str
        The original line from the requirements file.
    line_number: int
        The line number the requirement was on in the requirements file.
    """

    _converter = t.Dict(
        {
            t.Key("package_name"): String(),
            t.Key("constraints"): t.List(CustomDependencyConstraint.schema),
            t.Key("line"): String(),
            t.Key("line_number"): Int(gt=0),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self, **kwargs: Any) -> None:
        self._set_values(**kwargs)

    def __repr__(self) -> str:
        return "{}(package_name={!r}, constraints={!r})".format(
            self.__class__.__name__,
            self.package_name,
            self.constraints,
        )

    def _set_values(  # pylint: disable=missing-function-docstring
        self,
        package_name: str,
        constraints: List[CustomDependencyConstraint],
        line: str,
        line_number: int,
    ) -> None:
        self.package_name = package_name
        # TODO: maybe CustomDependencyConstraint should be implemented as TypedDict
        self.constraints = [CustomDependencyConstraint(**c) for c in constraints]  # type: ignore[arg-type]
        self.line = line
        self.line_number = line_number


class CustomModelVersion(APIObject):
    """A version of a DataRobot custom model.

    .. versionadded:: v2.21

    Attributes
    ----------
    id: str
        The ID of the custom model version.
    custom_model_id: str
        The ID of the custom model.
    version_minor: int
        A minor version number of the custom model version.
    version_major: int
        A major version number of the custom model version.
    is_frozen: bool
        A flag if the custom model version is frozen.
    items: List[CustomModelFileItem]
        A list of file items attached to the custom model version.
    base_environment_id: str
        The ID of the environment to use with the model.
    base_environment_version_id: str
        The ID of the environment version to use with the model.
    label: str, optional
        A short human readable string to label the version.
    description: str, optional
        The custom model version description.
    created_at: str, optional
        ISO-8601 formatted timestamp of when the version was created.
    dependencies: List[CustomDependency]
        The parsed dependencies of the custom model version if the
        version has a valid requirements.txt file.
    network_egress_policy: datarobot.NETWORK_EGRESS_POLICY, optional
        Determines whether the given custom model is isolated, or can access the public network.
        Values: [`datarobot.NETWORK_EGRESS_POLICY.NONE`, `datarobot.NETWORK_EGRESS_POLICY.DR_API_ACCESS`,
        `datarobot.NETWORK_EGRESS_POLICY.PUBLIC`].
        Note: `datarobot.NETWORK_EGRESS_POLICY.DR_API_ACCESS` value is only supported by the SaaS (cloud) environment.
    maximum_memory: int, optional
        The maximum memory that might be allocated by the custom-model.
        If exceeded, the custom-model will be killed by k8s.
    replicas: int, optional
        A fixed number of replicas that will be deployed in the cluster.
    required_metadata_values: List[RequiredMetadataValue]
        Additional parameters required by the execution environment. The required keys are
        defined by the fieldNames in the base environment's requiredMetadataKeys.
    training_data: TrainingData, optional
        The information about the training data assigned to the model version.
    holdout_data: HoldoutData, optional
        The information about the holdout data assigned to the model version.
    """

    _path = "customModels/{}/versions/"

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("custom_model_id"): String(),
            t.Key("version_minor"): Int(),
            t.Key("version_major"): Int(),
            t.Key("is_frozen"): t.Bool(),
            t.Key("items"): t.List(CustomModelFileItem.schema),
            # base_environment_id will be required once dependency management is enabled by default
            # in 6.2, but for backwards compatibility, this should be optional
            t.Key("base_environment_id", optional=True): String(),
            t.Key("base_environment_version_id", optional=True): String(),
            t.Key("label", optional=True): t.Or(String(max_length=50, allow_blank=True), t.Null()),
            t.Key("description", optional=True): t.Or(
                String(max_length=10000, allow_blank=True), t.Null()
            ),
            t.Key("created", optional=True) >> "created_at": String(),
            t.Key("dependencies", optional=True): t.List(CustomDependency.schema),
            t.Key("network_egress_policy", optional=True): t.Enum(*NETWORK_EGRESS_POLICY.ALL),
            t.Key("maximum_memory", optional=True): Int(),
            t.Key("replicas", optional=True): Int(),
            t.Key("required_metadata_values", optional=True): t.List(RequiredMetadataValue.schema),
            t.Key("training_data", optional=True): t.Or(TrainingData.schema, t.Null()),
            t.Key("holdout_data", optional=True): t.Or(HoldoutData.schema, t.Null()),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self, **kwargs: Any) -> None:
        self._set_values(**kwargs)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.label or self.id!r})"

    def _set_values(  # pylint: disable=missing-function-docstring
        self,
        id: str,
        custom_model_id: str,
        version_minor: int,
        version_major: int,
        is_frozen: bool,
        items: List[CustomModelFileItem],
        base_environment_id: Optional[str] = None,
        base_environment_version_id: Optional[str] = None,
        label: Optional[str] = None,
        description: Optional[str] = None,
        created_at: Optional[str] = None,
        dependencies: Optional[List[CustomDependency]] = None,
        network_egress_policy: Optional[NETWORK_EGRESS_POLICY] = None,
        maximum_memory: Optional[int] = None,
        replicas: Optional[int] = None,
        required_metadata_values: Optional[List[RequiredMetadataValue]] = None,
        training_data: Optional[Mapping[str, Any]] = None,
        holdout_data: Optional[Mapping[str, Any]] = None,
    ) -> None:
        self.id = id
        self.custom_model_id = custom_model_id
        self.version_minor = version_minor
        self.version_major = version_major
        self.is_frozen = is_frozen
        self.items = [CustomModelFileItem(**item) for item in items]  # type: ignore[arg-type]
        self.base_environment_id = base_environment_id
        self.base_environment_version_id = base_environment_version_id
        self.label = label
        self.description = description
        self.created_at = created_at
        self.dependencies = [CustomDependency(**dep) for dep in dependencies or []]  # type: ignore[arg-type]
        self.network_egress_policy = network_egress_policy
        self.maximum_memory = maximum_memory
        self.replicas = replicas
        self.required_metadata_values = (
            [RequiredMetadataValue.from_server_data(val) for val in required_metadata_values]  # type: ignore[arg-type]
            if required_metadata_values
            else None
        )
        self.training_data = TrainingData(**training_data) if training_data else None
        self.holdout_data = HoldoutData(**holdout_data) if holdout_data else None

    @classmethod
    def from_server_data(
        cls, data: ServerDataType, keep_attrs: Optional[Iterable[str]] = None
    ) -> CustomModelVersion:
        initial = super().from_server_data(data, keep_attrs)
        # from_server_data will make the keys in requiredMetadata lowercase,
        # which is not OK. we need to preserve case
        initial.required_metadata = data.get("requiredMetadata")  # type: ignore[union-attr]
        return initial

    @classmethod
    def create_clean(
        cls,
        custom_model_id: str,
        base_environment_id: str,
        is_major_update: bool = True,
        folder_path: Optional[str] = None,
        files: Optional[List[Tuple[str, str]]] = None,
        network_egress_policy: Optional[NETWORK_EGRESS_POLICY] = None,
        maximum_memory: Optional[int] = None,
        replicas: Optional[int] = None,
        required_metadata_values: Optional[List[RequiredMetadataValue]] = None,
        training_dataset_id: Optional[str] = None,
        partition_column: Optional[str] = None,
        holdout_dataset_id: Optional[str] = None,
        keep_training_holdout_data: Optional[bool] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> CustomModelVersion:
        """Create a custom model version without files from previous versions.

           Create a version with training or holdout data:
           If training/holdout data related parameters are provided,
           the training data is assigned asynchronously.
           In this case:
           * if max_wait is not None, the function returns once the job is finished.
           * if max_wait is None, the function returns immediately. Progress can be polled by the user (see examples).

           If training data assignment fails, new version is still created,
           but it is not allowed to create a model package for the model version and to deploy it.
           To check for training data assignment error, check version.training_data.assignment_error["message"].

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            The ID of the custom model.
        base_environment_id: str
            The ID of the base environment to use with the custom model version.
        is_major_update: bool
            The flag defining if a custom model version will be a minor or a major version.
            Default to `True`
        folder_path: str, optional
            The path to a folder containing files to be uploaded.
            Each file in the folder is uploaded under path relative to a folder path.
        files: list, optional
            The list of tuples, where values in each tuple are the local filesystem path and
            the path the file should be placed in the model.
            If the list is of strings, then basenames will be used for tuples.
            Example:
            [("/home/user/Documents/myModel/file1.txt", "file1.txt"),
            ("/home/user/Documents/myModel/folder/file2.txt", "folder/file2.txt")]
            or
            ["/home/user/Documents/myModel/file1.txt",
            "/home/user/Documents/myModel/folder/file2.txt"]
        network_egress_policy: datarobot.NETWORK_EGRESS_POLICY, optional
            Determines whether the given custom model is isolated, or can access the public network.
            Values: [`datarobot.NETWORK_EGRESS_POLICY.NONE`, `datarobot.NETWORK_EGRESS_POLICY.DR_API_ACCESS`,
            `datarobot.NETWORK_EGRESS_POLICY.PUBLIC`].
            Note: `datarobot.NETWORK_EGRESS_POLICY.DR_API_ACCESS` value
            is only supported by the SaaS (cloud) environment.
        maximum_memory: int, optional
            The maximum memory that might be allocated by the custom-model.
            If exceeded, the custom-model will be killed by k8s.
        replicas: int, optional
            A fixed number of replicas that will be deployed in the cluster.
        required_metadata_values: List[RequiredMetadataValue]
            Additional parameters required by the execution environment. The required keys are
            defined by the fieldNames in the base environment's requiredMetadataKeys.
        training_dataset_id: str, optional
            The ID of the training dataset to assign to the custom model.
        partition_column: str, optional
            Name of a partition column in a training dataset assigned to the custom model.
            Can only be assigned for structured models.
        holdout_dataset_id: str, optional
            The ID of the holdout dataset to assign to the custom model.
            Can only be assigned for unstructured models.
        keep_training_holdout_data: bool, optional
            If the version should inherit training and holdout data from the previous version.
            Defaults to True.
            This field is only applicable if the model has training data for versions enabled,
            otherwise the field value will be ignored.
        max_wait: int, optional
            Max time to wait for training data assignment.
            If set to None - method will return without waiting.
            Defaults to 10 minutes.

        Returns
        -------
        CustomModelVersion
            Created custom model version.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        datarobot.errors.InvalidUsageError
            If wrong parameters are provided.
        datarobot.errors.TrainingDataAssignmentError
            If training data assignment fails.

        Examples
        --------
        Create a version with blocking (default max_wait=600) training data assignment:

        .. code-block:: python

            import datarobot as dr
            from datarobot.errors import TrainingDataAssignmentError

            dr.Client(token=my_token, endpoint=endpoint)

            try:
                version = dr.CustomModelVersion.create_from_previous(
                    custom_model_id="6444482e5583f6ee2e572265",
                    base_environment_id="642209acc563893014a41e24",
                    training_dataset_id="6421f2149a4f9b1bec6ad6dd",
                )
            except TrainingDataAssignmentError as e:
                print(e)

        Create a version with non-blocking training data assignment:

        .. code-block:: python

            import datarobot as dr

            dr.Client(token=my_token, endpoint=endpoint)

            version = dr.CustomModelVersion.create_from_previous(
                custom_model_id="6444482e5583f6ee2e572265",
                base_environment_id="642209acc563893014a41e24",
                training_dataset_id="6421f2149a4f9b1bec6ad6dd",
                max_wait=None,
            )

            while version.training_data.assignment_in_progress:
                time.sleep(10)
                version.refresh()
            if version.training_data.assignment_error:
                print(version.training_data.assignment_error["message"])
        """
        if files and not isinstance(files[0], tuple) and isinstance(files[0], str):  # type: ignore[unreachable]
            files = [(filename, os.path.basename(filename)) for filename in files]  # type: ignore[unreachable]
        return cls._create(
            "post",
            custom_model_id,
            is_major_update,
            base_environment_id,
            folder_path,
            files,
            extra_upload_data=None,
            network_egress_policy=network_egress_policy,
            maximum_memory=maximum_memory,
            replicas=replicas,
            required_metadata_values=required_metadata_values,
            training_dataset_id=training_dataset_id,
            partition_column=partition_column,
            holdout_dataset_id=holdout_dataset_id,
            keep_training_holdout_data=keep_training_holdout_data,
            max_wait=max_wait,
        )

    @classmethod
    def create_from_previous(
        cls,
        custom_model_id: str,
        base_environment_id: str,
        is_major_update: bool = True,
        folder_path: Optional[str] = None,
        files: Optional[List[Tuple[str, str]]] = None,
        files_to_delete: Optional[List[str]] = None,
        network_egress_policy: Optional[NETWORK_EGRESS_POLICY] = None,
        maximum_memory: Optional[int] = None,
        replicas: Optional[int] = None,
        required_metadata_values: Optional[List[RequiredMetadataValue]] = None,
        training_dataset_id: Optional[str] = None,
        partition_column: Optional[str] = None,
        holdout_dataset_id: Optional[str] = None,
        keep_training_holdout_data: Optional[bool] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
    ) -> CustomModelVersion:
        """Create a custom model version containing files from a previous version.

           Create a version with training/holdout data:
           If training/holdout data related parameters are provided,
           the training data is assigned asynchronously.
           In this case:
           * if max_wait is not None, function returns once job is finished.
           * if max_wait is None, function returns immediately, progress can be polled by the user, see examples.

           If training data assignment fails, new version is still created,
           but it is not allowed to create a model package for the model version and to deploy it.
           To check for training data assignment error, check version.training_data.assignment_error["message"].


        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            The ID of the custom model.
        base_environment_id: str
            The ID of the base environment to use with the custom model version.
        is_major_update: bool, optional
            The flag defining if a custom model version will be a minor or a major version.
            Defaults to `True`.
        folder_path: str, optional
            The path to a folder containing files to be uploaded.
            Each file in the folder is uploaded under path relative to a folder path.
        files: list, optional
            The list of tuples, where values in each tuple are the local filesystem path and
            the path the file should be placed in the model.
            If list is of strings, then basenames will be used for tuples
            Example:
            [("/home/user/Documents/myModel/file1.txt", "file1.txt"),
            ("/home/user/Documents/myModel/folder/file2.txt", "folder/file2.txt")]
            or
            ["/home/user/Documents/myModel/file1.txt",
            "/home/user/Documents/myModel/folder/file2.txt"]
        files_to_delete: list, optional
            The list of a file items ids to be deleted.
            Example: ["5ea95f7a4024030aba48e4f9", "5ea6b5da402403181895cc51"]
        network_egress_policy: datarobot.NETWORK_EGRESS_POLICY, optional
            Determines whether the given custom model is isolated, or can access the public network.
            Values: [`datarobot.NETWORK_EGRESS_POLICY.NONE`, `datarobot.NETWORK_EGRESS_POLICY.DR_API_ACCESS`,
            `datarobot.NETWORK_EGRESS_POLICY.PUBLIC`].
            Note: `datarobot.NETWORK_EGRESS_POLICY.DR_API_ACCESS` value
            is only supported by the SaaS (cloud) environment.
        maximum_memory: int, optional
            The maximum memory that might be allocated by the custom-model.
            If exceeded, the custom-model will be killed by k8s
        replicas: int, optional
            A fixed number of replicas that will be deployed in the cluster
        required_metadata_values: List[RequiredMetadataValue]
            Additional parameters required by the execution environment. The required keys are
            defined by the fieldNames in the base environment's requiredMetadataKeys.
        training_dataset_id: str, optional
            The ID of the training dataset to assign to the custom model.
        partition_column: str, optional
            Name of a partition column in a training dataset assigned to the custom model.
            Can only be assigned for structured models.
        holdout_dataset_id: str, optional
            The ID of the holdout dataset to assign to the custom model.
            Can only be assigned for unstructured models.
        keep_training_holdout_data: bool, optional
            If the version should inherit training and holdout data from the previous version.
            Defaults to True.
            This field is only applicable if the model has training data for versions enabled,
            otherwise the field value will be ignored.
        max_wait: int, optional
            Max time to wait for training data assignment.
            If set to None - method will return without waiting.
            Defaults to 10 minutes.

        Returns
        -------
        CustomModelVersion
            created custom model version

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        datarobot.errors.InvalidUsageError
            If wrong parameters are provided.
        datarobot.errors.TrainingDataAssignmentError
            If training data assignment fails.

        Examples
        --------
        Create a version with blocking (default max_wait=600) training data assignment:

        .. code-block:: python

            import datarobot as dr
            from datarobot.errors import TrainingDataAssignmentError

            dr.Client(token=my_token, endpoint=endpoint)

            try:
                version = dr.CustomModelVersion.create_from_previous(
                    custom_model_id="6444482e5583f6ee2e572265",
                    base_environment_id="642209acc563893014a41e24",
                    training_dataset_id="6421f2149a4f9b1bec6ad6dd",
                )
            except TrainingDataAssignmentError as e:
                print(e)

        Create a version with non-blocking training data assignment:

        .. code-block:: python

            import datarobot as dr

            dr.Client(token=my_token, endpoint=endpoint)

            version = dr.CustomModelVersion.create_from_previous(
                custom_model_id="6444482e5583f6ee2e572265",
                base_environment_id="642209acc563893014a41e24",
                training_dataset_id="6421f2149a4f9b1bec6ad6dd",
                max_wait=None,
            )

            while version.training_data.assignment_in_progress:
                time.sleep(10)
                version.refresh()
            if version.training_data.assignment_error:
                print(version.training_data.assignment_error["message"])

        """
        if files and not isinstance(files[0], tuple) and isinstance(files[0], str):  # type: ignore[unreachable]
            files = [(filename, os.path.basename(filename)) for filename in files]  # type: ignore[unreachable]
        if files_to_delete:
            upload_data = [("filesToDelete", file_id) for file_id in files_to_delete]
        else:
            upload_data = None
        return cls._create(
            "patch",
            custom_model_id,
            is_major_update,
            base_environment_id,
            folder_path,
            files,
            upload_data,
            network_egress_policy,
            maximum_memory,
            replicas,
            required_metadata_values=required_metadata_values,
            training_dataset_id=training_dataset_id,
            partition_column=partition_column,
            holdout_dataset_id=holdout_dataset_id,
            keep_training_holdout_data=keep_training_holdout_data,
            max_wait=max_wait,
        )

    @classmethod
    def _create(
        cls,
        method: str,
        custom_model_id: str,
        is_major_update: bool,
        base_environment_id: str,
        folder_path: Optional[str],
        files: Optional[List[Tuple[str, str]]],
        extra_upload_data: Optional[List[Tuple[str, Any]]],
        network_egress_policy: Optional[NETWORK_EGRESS_POLICY],
        maximum_memory: Optional[int],
        replicas: Optional[int],
        required_metadata_values: Optional[List[RequiredMetadataValue]],
        training_dataset_id: Optional[str],
        partition_column: Optional[str],
        holdout_dataset_id: Optional[str],
        keep_training_holdout_data: Optional[bool],
        max_wait: Optional[int],
    ) -> CustomModelVersion:
        # TODO: pass model object
        """Create a custom model version"""

        def _wait_for_training_data_assignment(version: CustomModelVersion) -> None:
            # This check is needed to make sure user explicitly passes new training data.
            # Checking only for `version.training_data.assignment_in_progress` is not enough as it is not known
            # whether this training data is new or copied from the previous version.
            # TODO: replace `not holdout_dataset_id` with `not model.is_unstructured_model_kind`
            #  once `model` entity is introduced in the API
            if training_dataset_id and max_wait and not holdout_dataset_id:
                start_time = time.time()
                while time.time() < start_time + max_wait:
                    version.refresh()
                    if (
                        not version.training_data
                        or not version.training_data.assignment_in_progress
                    ):
                        break
                    time.sleep(5)

                if version.training_data and version.training_data.assignment_error:
                    raise TrainingDataAssignmentError(
                        version.custom_model_id,
                        version.id,
                        version.training_data.assignment_error["message"],
                    )

        url = cls._path.format(custom_model_id)

        with contextlib.ExitStack() as stack:
            # TODO: add training data params checks depending on structured/unstructured model kind.
            if not training_dataset_id and (holdout_dataset_id or partition_column):
                raise InvalidUsageError(
                    "It is not allowed to provide holdout data or partition column without training data."
                )

            if holdout_dataset_id and partition_column:
                raise InvalidUsageError(
                    "It is not allowed to provide holdout_dataset_id and partition_column at the same time. "
                    "For regular(structured) models you can provide training_dataset_id and partition_column. "
                    "For unstructured models you can provide training_dataset_id and holdout_dataset_id."
                )

            if keep_training_holdout_data and (training_dataset_id or holdout_dataset_id):
                raise InvalidUsageError(
                    "It is not allowed to keep existing training/holdout data and to provide a new ones."
                )

            upload_data: List[Tuple[str, Any]] = [
                ("isMajorUpdate", str(is_major_update)),
                ("baseEnvironmentId", base_environment_id),
            ]

            if folder_path:
                for root_path, _, file_paths in os.walk(folder_path):
                    for path in file_paths:
                        file_path = os.path.join(root_path, path)
                        file = stack.enter_context(open(file_path, "rb"))

                        upload_data.append(("file", (os.path.basename(file_path), file)))
                        upload_data.append(("filePath", os.path.relpath(file_path, folder_path)))

            if files:
                for file_path, upload_file_path in files:
                    file = stack.enter_context(open(file_path, "rb"))

                    upload_data.append(("file", (os.path.basename(upload_file_path), file)))
                    upload_data.append(("filePath", upload_file_path))

            if extra_upload_data:
                upload_data += extra_upload_data

            if network_egress_policy:
                upload_data.append(("networkEgressPolicy", network_egress_policy))

            if maximum_memory:
                upload_data.append(("maximumMemory", str(maximum_memory)))

            if replicas:
                upload_data.append(("replicas", str(replicas)))

            if required_metadata_values is not None:
                upload_data.append(
                    (
                        "requiredMetadataValues",
                        json.dumps(
                            [
                                {camelize(k): v for k, v in val.to_dict().items()}
                                for val in required_metadata_values
                            ]
                        ),
                    )
                )

            if training_dataset_id:
                keep_training_holdout_data = False

                td_payload = {
                    "datasetId": training_dataset_id,
                }
                upload_data.append(("trainingData", json.dumps(td_payload)))

                # If holdout data is not provided don't include it in the payload with None values.
                # There should be more checks added in the API.
                hd_payload = {}
                if holdout_dataset_id:
                    hd_payload["datasetId"] = holdout_dataset_id
                if partition_column:
                    hd_payload["partitionColumn"] = partition_column

                if len(hd_payload):
                    upload_data.append(("holdoutData", json.dumps(hd_payload)))
            if keep_training_holdout_data is not None:
                upload_data.append(("keepTrainingHoldoutData", str(keep_training_holdout_data)))

            encoder = MultipartEncoder(fields=upload_data)
            headers = {"Content-Type": encoder.content_type}
            response = cls._client.request(method, url, data=encoder, headers=headers)
            new_version = cls.from_server_data(response.json())

            _wait_for_training_data_assignment(new_version)

        return new_version

    @classmethod
    def list(cls, custom_model_id: str) -> List[CustomModelVersion]:
        """List custom model versions.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            The ID of the custom model.

        Returns
        -------
        List[CustomModelVersion]
            A list of custom model versions.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        url = cls._path.format(custom_model_id)
        data = unpaginate(url, None, cls._client)
        return [cls.from_server_data(item) for item in data]

    @classmethod
    def get(cls, custom_model_id: str, custom_model_version_id: str) -> CustomModelVersion:
        """Get custom model version by id.

        .. versionadded:: v2.21

        Parameters
        ----------
        custom_model_id: str
            The ID of the custom model.
        custom_model_version_id: str
            The id of the custom model version to retrieve.

        Returns
        -------
        CustomModelVersion
            Retrieved custom model version.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        url = cls._path.format(custom_model_id)
        path = f"{url}{custom_model_version_id}/"
        return cls.from_location(path)

    def download(self, file_path: str) -> None:
        """Download custom model version.

        .. versionadded:: v2.21

        Parameters
        ----------
        file_path: str
            Path to create a file with custom model version content.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        url = self._path.format(self.custom_model_id)
        path = f"{url}{self.id}/download/"

        response = self._client.get(path)
        with open(file_path, "wb") as f:
            f.write(response.content)

    def update(
        self,
        description: Optional[str] = None,
        required_metadata_values: Optional[List[RequiredMetadataValue]] = None,
    ) -> None:
        """Update custom model version properties.

        .. versionadded:: v2.21

        Parameters
        ----------
        description: str, optional
            New custom model version description.
        required_metadata_values: List[RequiredMetadataValue], optional
            Additional parameters required by the execution environment. The required keys are
            defined by the fieldNames in the base environment's requiredMetadataKeys.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        payload: Dict[str, Any] = {}
        if description:
            payload.update({"description": description})

        if required_metadata_values is not None:
            payload.update(
                {"requiredMetadataValues": [val.to_dict() for val in required_metadata_values]}
            )

        url = self._path.format(self.custom_model_id)
        path = f"{url}{self.id}/"

        response = self._client.patch(path, data=payload)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))  # type: ignore[no-untyped-call]
        # _safe_data will make the keys in requiredMetadata lowercase,
        # which is not OK. we need to preserve case
        self.required_metadata = data.get(  # pylint: disable=attribute-defined-outside-init
            "requiredMetadata"
        )

    def refresh(self) -> None:
        """Update custom model version with the latest data from server.

        .. versionadded:: v2.21

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        url = self._path.format(self.custom_model_id)
        path = f"{url}{self.id}/"

        response = self._client.get(path)

        data = response.json()
        self._set_values(**self._safe_data(data, do_recursive=True))  # type: ignore[no-untyped-call]

    def get_feature_impact(self, with_metadata: bool = False) -> List[Dict[str, Any]]:
        """Get custom model feature impact.

        .. versionadded:: v2.23

        Parameters
        ----------
        with_metadata : bool
            The flag indicating if the result should include the metadata as well.

        Returns
        -------
        feature_impacts : list of dict
           The feature impact data. Each item is a dict with the keys 'featureName',
           'impactNormalized', and 'impactUnnormalized', and 'redundantWith'.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """
        url = self._path.format(self.custom_model_id)
        path = f"{url}{self.id}/featureImpact/"
        data = self._client.get(path).json()
        data = custom_model_feature_impact_trafaret.check(data)
        ret = filter_feature_impact_result(data, with_metadata=with_metadata)  # type: ignore[no-untyped-call]
        return ret  # type: ignore[no-any-return]

    def calculate_feature_impact(self, max_wait: int = DEFAULT_MAX_WAIT) -> None:
        """Calculate custom model feature impact.

        .. versionadded:: v2.23

        Parameters
        ----------
        max_wait: int, optional
            Max time to wait for feature impact calculation.
            If set to None - method will return without waiting.
            Defaults to 10 min

        Raises
        ------
        datarobot.errors.ClientError
            if the server responded with 4xx status
        datarobot.errors.ServerError
            if the server responded with 5xx status
        """
        url = self._path.format(self.custom_model_id)
        path = f"{url}{self.id}/featureImpact/"
        response = self._client.post(path)

        if max_wait is not None:
            wait_for_async_resolution(self._client, response.headers["Location"], max_wait)


class CustomModelVersionConversion(APIObject):
    """A conversion of a DataRobot custom model version.

    .. versionadded:: v2.27

    Attributes
    ----------
    id: str
        The ID of the custom model version conversion.
    custom_model_version_id: str
        The ID of the custom model version.
    created: str
        ISO-8601 timestamp of when the custom model conversion created.
    main_program_item_id: str or None
        The ID of the main program item.
    log_message: str or None
        The conversion output log message.
    generated_metadata: dict or None
        The dict contains two items: 'outputDataset' & 'outputColumns'.
    conversion_succeeded: bool
        Whether the conversion succeeded or not.
    conversion_in_progress: bool
        Whether a given conversion is in progress or not.
    should_stop: bool
        Whether the user asked to stop a conversion.
    """

    _path = "customModels/{}/versions/{}/conversions/"

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("custom_model_version_id"): String(),
            t.Key("created"): String(),
            t.Key("main_program_item_id", optional=True): String(),
            t.Key("log_message", optional=True): t.Or(String(), t.Null()),
            t.Key("generated_metadata", optional=True): t.Dict(
                {
                    t.Key("output_datasets"): t.List(t.String, min_length=0, max_length=50),
                    t.Key("output_columns"): t.List(
                        t.List(t.String, min_length=0, max_length=1024), min_length=1, max_length=50
                    ),
                }
            ),
            t.Key("conversion_succeeded", optional=True): t.Bool() | t.Null(),
            t.Key("conversion_in_progress", optional=True): t.Bool() | t.Null(),
            t.Key("should_stop", optional=True): t.Bool(),
        }
    ).ignore_extra("*")

    schema = _converter

    def __init__(self, **kwargs: Any) -> None:
        self._set_values(**kwargs)
        self.custom_model_id = None

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.id!r})"

    def _set_values(  # pylint: disable=missing-function-docstring
        self,
        id: str,
        custom_model_version_id: str,
        created: str,
        main_program_item_id: Optional[str] = None,
        log_message: Optional[str] = None,
        generated_metadata: Optional[Dict[str, Union[List[str], List[List[str]]]]] = None,
        conversion_succeeded: Optional[bool] = None,
        conversion_in_progress: Optional[bool] = None,
        should_stop: Optional[bool] = None,
    ) -> None:
        self.id = id
        self.custom_model_version_id = custom_model_version_id
        self.created = created
        self.main_program_item_id = main_program_item_id
        self.log_message = log_message
        self.generated_metadata = generated_metadata
        self.conversion_succeeded = conversion_succeeded
        self.conversion_in_progress = conversion_in_progress
        self.should_stop = should_stop

    @classmethod
    def run_conversion(
        cls,
        custom_model_id: str,
        custom_model_version_id: str,
        main_program_item_id: str,
        max_wait: Optional[int] = None,
    ) -> str:
        """Initiate a new custom model version conversion.

        Parameters
        ----------
        custom_model_id : str
            The associated custom model ID.
        custom_model_version_id : str
            The associated custom model version ID.
        main_program_item_id : str
            The selected main program item ID. This should be one of the SAS items in the
            associated custom model version.
        max_wait: int or None
            Max wait time in seconds. If None, then don't wait.

        Returns
        -------
        conversion_id : str
            The ID of the newly created conversion entity.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """

        base_conversion_url = cls._path.format(custom_model_id, custom_model_version_id)
        payload = {"mainProgramItemId": main_program_item_id}
        response = cls._client.post(base_conversion_url, json=payload)
        if max_wait is not None:
            wait_for_async_resolution(cls._client, response.headers["Location"], max_wait)
        return cast(str, response.json()["conversionId"])

    @classmethod
    def stop_conversion(
        cls, custom_model_id: str, custom_model_version_id: str, conversion_id: str
    ) -> Response:
        """
        Stop a conversion that is in progress.

        Parameters
        ----------
        custom_model_id : str
            The ID of the associated custom model.
        custom_model_version_id : str
            The ID of the associated custom model version.
        conversion_id :
            THe ID of a conversion that is in-progress.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """

        conversion = cls.get(custom_model_id, custom_model_version_id, conversion_id)
        if not conversion.conversion_in_progress:
            raise InvalidUsageError("You may only stop a conversion that is in progress.")

        base_conversion_url = cls._path.format(custom_model_id, custom_model_version_id)
        conversion_entity_url = "{}{}/".format(base_conversion_url, conversion_id)
        return cast(Response, cls._client.delete(conversion_entity_url))

    @classmethod
    def get(
        cls, custom_model_id: str, custom_model_version_id: str, conversion_id: str
    ) -> CustomModelVersionConversion:
        """Get custom model version conversion by id.

        .. versionadded:: v2.27

        Parameters
        ----------
        custom_model_id: str
            The ID of the custom model.
        custom_model_version_id: str
            The ID of the custom model version.
        conversion_id: str
            The ID of the conversion to retrieve.

        Returns
        -------
        CustomModelVersionConversion
            Retrieved custom model version conversion.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """

        base_conversion_url = cls._path.format(custom_model_id, custom_model_version_id)
        conversion_entity_url = "{}{}/".format(base_conversion_url, conversion_id)
        return cls.from_location(conversion_entity_url)

    @classmethod
    def get_latest(
        cls, custom_model_id: str, custom_model_version_id: str
    ) -> Optional[CustomModelVersionConversion]:
        """Get latest custom model version conversion for a given custom model version.

        .. versionadded:: v2.27

        Parameters
        ----------
        custom_model_id: str
            The ID of the custom model.
        custom_model_version_id: str
            The ID of the custom model version.

        Returns
        -------
        CustomModelVersionConversion or None
            Retrieved latest conversion for a given custom model version.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """

        base_conversion_url = cls._path.format(custom_model_id, custom_model_version_id)
        data = unpaginate(base_conversion_url, {"isLatest": True}, cls._client)
        return cls.from_server_data(next(data))

    @classmethod
    def list(
        cls, custom_model_id: str, custom_model_version_id: str
    ) -> List[CustomModelVersionConversion]:
        """Get custom model version conversions list per custom model version.

        .. versionadded:: v2.27

        Parameters
        ----------
        custom_model_id: str
            The ID of the custom model.
        custom_model_version_id: str
            The ID of the custom model version.

        Returns
        -------
        List[CustomModelVersionConversion]
            Retrieved conversions for a given custom model version.

        Raises
        ------
        datarobot.errors.ClientError
            If the server responded with 4xx status.
        datarobot.errors.ServerError
            If the server responded with 5xx status.
        """

        base_conversion_url = cls._path.format(custom_model_id, custom_model_version_id)
        data = unpaginate(base_conversion_url, None, cls._client)
        return [cls.from_server_data(item) for item in data]



#
# Copyright 2021-2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
import re

DESCRIPTION_TEMPLATE = """
About {package_name}
============================
.. image:: https://img.shields.io/pypi/v/{package_name}.svg
   :target: {pypi_url_target}
.. image:: https://img.shields.io/pypi/pyversions/{package_name}.svg
.. image:: https://img.shields.io/pypi/status/{package_name}.svg

DataRobot is a client library for working with the `DataRobot`_ platform API. {extra_desc}

This package is released under the terms of the DataRobot Tool and Utility Agreement, which
can be found on our `Legal`_ page, along with our privacy policy and more.

Installation
=========================
Python {python_versions} are supported.
You must have a datarobot account.

::

   $ pip install {pip_package_name}

Usage
=========================
The library will look for a config file `~/.config/datarobot/drconfig.yaml` by default.
This is an example of what that config file should look like.

::

   token: your_token
   endpoint: https://app.datarobot.com/api/v2

Alternatively a global client can be set in the code.

::

   import datarobot as dr
   dr.Client(token='your_token', endpoint='https://app.datarobot.com/api/v2')

Alternatively environment variables can be used.

::

   export DATAROBOT_API_TOKEN='your_token'
   export DATAROBOT_ENDPOINT='https://app.datarobot.com/api/v2'

See `documentation`_ for example usage after configuring.

Helpful links
=========================
- `API quickstart guide <https://docs.datarobot.com/en/docs/api/api-quickstart/index.html>`_
- `Code examples <https://docs.datarobot.com/en/docs/api/guide/python/index.html>`_
- `Common use cases <https://docs.datarobot.com/en/docs/api/guide/common-case/index.html>`_

Bug Reporting and Q&A
=========================
Create an account on `DataRobot Community <https://community.datarobot.com/>`_ and post your questions and bug reports
`here <https://community.datarobot.com/t5/forums/postpage/board-id/datarobot-api>`_.

.. _datarobot: https://datarobot.com
.. _documentation: {docs_link}
.. _legal: https://www.datarobot.com/legal/
"""


DEFAULT_CLASSIFIERS = [
    "Development Status :: 5 - Production/Stable",
    "Intended Audience :: Developers",
    "Topic :: Scientific/Engineering :: Artificial Intelligence",
    "License :: Other/Proprietary License",
    "Operating System :: OS Independent",
    "Programming Language :: Python :: 3",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Programming Language :: Python :: 3.11",
]


with open("datarobot/_version.py") as fd:
    version_search = re.search(r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', fd.read(), re.MULTILINE)
    if not version_search:
        raise RuntimeError("Cannot find version information")
    version = version_search.group(1)

if not version:
    raise RuntimeError("Cannot find version information")

_mypy_require = [
    "mypy==1.0.0",
    "types-PyYAML==6.0.12",
    "types-python-dateutil==2.8.19",
    "types-pytz==2022.2.1.0",
    "types-requests==2.28.11",
    "types-urllib3==1.26.25",
]

images_require = ["Pillow==9.2.0"]

lint_require = (
    [
        "black==22.8.0",
        "black[jupyter]==22.8.0",
        "isort==5.10.1",
        "flake8==5.0.4",
        "pylint==2.15.0",
    ]
    + _mypy_require
    + images_require
)

tests_require = [
    "mock==3.0.5",
    "pytest==7.1.2",
    "pytest-cov",
    "responses==0.21",
    "pytest-asyncio",
] + images_require

dev_require = (
    tests_require
    + lint_require
    + images_require
    + [
        "Sphinx==4.3.2",
        "sphinx_rtd_theme==1.0.0",
        "nbsphinx==0.8.9",
        "numpydoc==1.4.0",
        "jupyter_contrib_nbextensions",
        "sphinx-autodoc-typehints==1.17.1",
    ]
)

example_require = [
    "jupyter<=5.0",
    "fredapi==0.4.0",
    "matplotlib>=2.1.0",
    "seaborn<=0.8",
    "scikit-learn<=0.18.2",
    "wordcloud<=1.3.1",
    "colour<=0.1.4",
]

release_require = ["zest.releaser[recommended]==6.22.0"]

# The None-valued kwargs should be updated by the caller
common_setup_kwargs = dict(
    name=None,
    version=None,
    description="This client library is designed to support the DataRobot API.",
    author="datarobot",
    author_email="support@datarobot.com",
    maintainer="datarobot",
    maintainer_email="info@datarobot.com",
    url="https://datarobot.com",
    project_urls={
        "Documentation": "https://datarobot-public-api-client.readthedocs-hosted.com/",
        "Changelog": "https://datarobot-public-api-client.readthedocs-hosted.com/page/CHANGES.html",
        "Bug Reporting": "https://community.datarobot.com/t5/forums/postpage/board-id/datarobot-api",
    },
    license="DataRobot Tool and Utility Agreement",
    packages=None,
    package_data={"datarobot": ["py.typed"]},
    python_requires=">=3.7",
    long_description=None,
    classifiers=None,
    install_requires=[
        "pandas>=0.15",
        "numpy",
        "pyyaml>=3.11",
        "requests>=2.28.1",
        "requests_toolbelt>=0.6",
        "trafaret>=0.7,<2.2,!=1.1.0",
        "urllib3>=1.23,<2.0.0",
        "typing-extensions>=4.3.0,<5",
        "mypy-extensions>=0.4.0,<2",
        "strenum>=0.4.15",
    ],
    extras_require={
        "dev": dev_require,
        "examples": example_require,
        "release": release_require,
        "lint": lint_require,
        "images": images_require,
        "test": tests_require,
    },
)



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Optional, TYPE_CHECKING

import trafaret as t

from datarobot._compat import Int, String
from datarobot.models.api_object import APIObject
from datarobot.utils import get_id_from_response

if TYPE_CHECKING:
    from datarobot.models.job import Job


class Ruleset(APIObject):
    """Represents an approximation of a model with DataRobot Prime

    Attributes
    ----------
    id : str
        the id of the ruleset
    rule_count : int
        the number of rules used to approximate the model
    score : float
        the validation score of the approximation
    project_id : str
        the project the approximation belongs to
    parent_model_id : str
        the model being approximated
    model_id : str or None
        the model using this ruleset (if it exists).  Will be None if no such model has been
        trained.

    """

    _converter = t.Dict(
        {
            t.Key("project_id"): String(),
            t.Key("parent_model_id"): String(),
            t.Key("model_id", optional=True): String(),
            t.Key("ruleset_id"): Int(),
            t.Key("rule_count"): Int(),
            t.Key("score"): t.Float(),
        }
    ).allow_extra("*")

    def __init__(
        self,
        project_id: str,
        parent_model_id: str,
        ruleset_id: int,
        rule_count: int,
        score: float,
        model_id: Optional[str] = None,
    ) -> None:
        self.id = ruleset_id
        self.rule_count = rule_count
        self.score = score
        self.project_id = project_id
        self.parent_model_id = parent_model_id
        self.model_id = model_id

    def __repr__(self) -> str:
        return f"Ruleset(rule_count={self.rule_count}, score={self.score})"

    def request_model(self) -> Job:
        """Request training for a model using this ruleset

        Training a model using a ruleset is a necessary prerequisite for being able to download
        the code for a ruleset.

        Returns
        -------
        job: Job
            the job fitting the new Prime model
        """
        from . import Job  # pylint: disable=import-outside-toplevel,cyclic-import

        if self.model_id:
            raise ValueError("Model already exists for ruleset")
        if not self.project_id:
            raise ValueError("Project ID needed in order to get ruleset model.")
        url = f"projects/{self.project_id}/primeModels/"
        data = {"parent_model_id": self.parent_model_id, "ruleset_id": self.id}
        response = self._client.post(url, data=data)
        job_id = get_id_from_response(response)
        return Job.get(self.project_id, job_id)



#
# Copyright 2023 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from contextvars import ContextVar
from dataclasses import dataclass, field
import functools
from typing import Callable, List, Optional, TYPE_CHECKING, TypeVar, Union

from trafaret import DataError
from typing_extensions import ParamSpec

from .errors import ClientError

if TYPE_CHECKING:
    from .models.use_cases.use_case import UseCase

DefaultUseCase = Optional[Union["UseCase", str, List[str], List["UseCase"]]]

# A single place to set the default since this will change in the future
ENABLE_API_CONSUMER_TRACKING_DEFAULT = True

P = ParamSpec("P")
T = TypeVar("T")


def env_to_bool(env_value: Union[str, int, bool]) -> bool:
    """
    A helper function to convert environment variable strings or ints into bools
    """
    return env_value not in ["no", "No", "0", "false", "False"]


def _init_context(func: Callable[P, T]) -> Callable[P, T]:
    """
    Ensures that context has been initialized before executing the decorated function.
    """

    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
        # ensure that context has been initialized (this is done as a side effect of creating the client)
        from .client import get_client  # pylint: disable=import-outside-toplevel, cyclic-import

        get_client()
        return func(*args, **kwargs)

    return wrapper


@dataclass
class _ContextGlobals:
    """
    Interface for initializing, accessing, and setting variables that should persist.

    This class should not be used directly; rather, reference `datarobot.context.Context`.

    Examples
    --------
        .. code-block:: python

            from datarobot.context import Context

            >>> def foobar(value: str) -> None:
            ...  print(f"hey {value}")
            ...  Context.use_case = value
            ...
            >>> foobar("Jude")
            hey Jude
            >>>
    """

    _use_case: DefaultUseCase = field(init=False, default=None)
    _enable_api_consumer_tracking: bool = field(
        init=False, default=ENABLE_API_CONSUMER_TRACKING_DEFAULT
    )
    _trace_context: Optional[str] = field(init=False, default=None)

    @property
    def use_case(self) -> DefaultUseCase:
        """Returns the default Use Case. If a string representing the Use Case ID
        was provided, lookup, cache, and return the actual entity."""
        return self.get_use_case()

    @use_case.setter
    def use_case(self, value: DefaultUseCase = None) -> None:
        self._use_case = value

    @_init_context
    def get_use_case(self, raw: bool = False) -> DefaultUseCase:
        """Returns the default Use Case. If a string representing the Use Case ID
        was provided and raw=False, lookup, cache, and return the actual entity.
        Otherwise, simply return the stored default Use Case in its current form"""
        if not raw:
            if isinstance(self._use_case, (str, list)):
                try:
                    from .models.use_cases.use_case import (  # pylint: disable=import-outside-toplevel
                        UseCase,
                    )

                    if isinstance(self._use_case, list):
                        if all(isinstance(use_case, str) for use_case in self._use_case):
                            self._use_case = [
                                UseCase.get(use_case_id=use_case_id)  # type: ignore[arg-type]
                                for use_case_id in self._use_case
                            ]
                    else:
                        self._use_case = UseCase.get(use_case_id=self._use_case)
                except (DataError, ClientError) as e:
                    raise ValueError("Current use case is invalid.", e)
        return self._use_case or None

    @property
    def enable_api_consumer_tracking(self) -> bool:
        return self._enable_api_consumer_tracking

    @enable_api_consumer_tracking.setter
    def enable_api_consumer_tracking(
        self, value: bool = ENABLE_API_CONSUMER_TRACKING_DEFAULT
    ) -> None:
        self._enable_api_consumer_tracking = value

    @property
    def trace_context(self) -> Optional[str]:
        return self._trace_context

    @trace_context.setter
    def trace_context(self, value: Optional[str] = None) -> None:
        self._trace_context = value


_context: ContextVar[_ContextGlobals] = ContextVar("current_context", default=_ContextGlobals())

Context = _context.get()



#
# Copyright 2021 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from datetime import datetime
from typing import cast, List, Optional, Tuple

from requests import Response
import trafaret as t
from typing_extensions import TypedDict

from datarobot._compat import String
from datarobot.enums import DEFAULT_MAX_WAIT, DocumentType, Locales
from datarobot.models.api_object import APIObject
from datarobot.utils import camelize, from_api, get_id_from_location, parse_time
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution


class DocumentOption(TypedDict):
    document_type: DocumentType
    locale: Locales


class AutomatedDocument(APIObject):
    """
    An :ref:`automated documentation <automated_documentation_overview>` object.

    .. versionadded:: v2.24

    Attributes
    ----------
    document_type : str or None
        Type of automated document. You can specify: ``MODEL_COMPLIANCE``, ``AUTOPILOT_SUMMARY``
        depending on your account settings. Required for document generation.
    entity_id : str or None
        ID of the entity to generate the document for. It can be model ID or project ID.
        Required for document generation.
    output_format : str or None
        Format of the generate document, either ``docx`` or ``html``.
        Required for document generation.
    locale : str or None
        Localization of the document, dependent on your account settings.
        Default setting is ``EN_US``.
    template_id : str or None
        Template ID to use for the document outline. Defaults to standard DataRobot template.
        See the documentation for :py:class:`ComplianceDocTemplate
        <datarobot.models.compliance_doc_template.ComplianceDocTemplate>` for more information.
    id :  str or None
        ID of the document. Required to download or delete a document.
    filepath : str or None
        Path to save a downloaded document to. Either include a file path and name or the file
        will be saved to the directory from which the script is launched.
    created_at :  datetime or None
        Document creation timestamp.

    """

    DEFAULT_BATCH_SIZE = 100

    _path = "automatedDocuments/"
    _model_compliance_initialization_path = "modelComplianceDocsInitializations/"

    _converter = t.Dict(
        {
            t.Key("id"): String(),
            t.Key("created_at"): parse_time,
            t.Key("entity_id"): String(),
            t.Key("output_format"): String(),
            t.Key("locale"): String(),
            t.Key("document_type"): String(),
            t.Key("template_id", optional=True): t.Or(String(), t.Null()),
        }
    ).allow_extra("*")

    def __init__(
        self,
        entity_id: Optional[str] = None,
        document_type: Optional[str] = None,
        output_format: Optional[str] = None,
        locale: Optional[str] = None,
        template_id: Optional[str] = None,
        id: Optional[str] = None,
        filepath: Optional[str] = None,
        created_at: Optional[datetime] = None,
    ) -> None:
        self.entity_id = entity_id
        self.document_type = document_type
        self.output_format = output_format
        self.locale = locale
        self.template_id = template_id
        self.id = id
        self.filepath = filepath
        self.created_at = created_at

    @classmethod
    def list_available_document_types(cls) -> List[DocumentOption]:
        """
        Get a list of all available document types and locales.

        Returns
        -------
        List of dicts

        Examples
        --------
        .. code-block:: python

            import datarobot as dr

            dr.Client(token=my_token, endpoint=endpoint)
            doc_types = dr.AutomatedDocument.list_available_document_types()

        """

        response = cls._client.get("automatedDocumentOptions/")
        return cast(List[DocumentOption], from_api(response.json()))

    @property
    def is_model_compliance_initialized(self) -> Tuple[bool, str]:
        """Check if model compliance documentation pre-processing is initialized.
        Model compliance documentation pre-processing must be initialized before
        generating documentation for a custom model.

        Returns
        -------
        Tuple of (boolean, string)
            - `boolean` flag is whether model compliance documentation pre-processing is initialized
            - `string` value is the initialization status
        """
        return self._model_compliance_initialization_status(
            f"{self._model_compliance_initialization_path}{self.entity_id}/"
        )

    def initialize_model_compliance(self) -> Tuple[bool, str]:
        """Initialize model compliance documentation pre-processing.
        Must be called before generating documentation for a custom model.

        Returns
        -------
        Tuple of (boolean, string)
            - `boolean` flag is whether model compliance documentation pre-processing is initialized
            - `string` value is the initialization status

        Examples
        --------
        .. code-block:: python

            import datarobot as dr

            dr.Client(token=my_token, endpoint=endpoint)

            # NOTE: entity_id is either a model id or a model package id
            doc = dr.AutomatedDocument(
                    document_type="MODEL_COMPLIANCE",
                    entity_id="6f50cdb77cc4f8d1560c3ed5",
                    output_format="docx",
                    locale="EN_US")

            doc.initialize_model_compliance()

        """

        response = self._client.post(
            f"{self._model_compliance_initialization_path}{self.entity_id}/"
        )
        location = wait_for_async_resolution(self._client, response.headers["Location"])
        return self._model_compliance_initialization_status(location)

    def generate(self, max_wait: int = DEFAULT_MAX_WAIT) -> Response:
        """Request generation of an automated document.

        Required attributes to request document generation: ``document_type``, ``entity_id``,
        and ``output_format``.

        Returns
        -------
        :class:`requests.models.Response`

        Examples
        --------
        .. code-block:: python

            import datarobot as dr

            dr.Client(token=my_token, endpoint=endpoint)

            doc = dr.AutomatedDocument(
                    document_type="MODEL_COMPLIANCE",
                    entity_id="6f50cdb77cc4f8d1560c3ed5",
                    output_format="docx",
                    locale="EN_US",
                    template_id="50efc9db8aff6c81a374aeec",
                    filepath="/Users/username/Documents/example.docx"
                    )

            doc.generate()
            doc.download()

        """
        payload = {
            "entity_id": self.entity_id,
            "document_type": self.document_type,
            "output_format": self.output_format,
            "locale": self.locale,
            "template_id": self.template_id,
        }
        payload = {key: val for key, val in payload.items() if val is not None}

        response = self._client.post(self._path, data=payload)
        location = wait_for_async_resolution(
            self._client, response.headers["Location"], max_wait=max_wait
        )
        self.id = get_id_from_location(location)

        return cast(Response, response)

    def download(self) -> Response:
        """
        Download a generated Automated Document.
        Document ID is required to download a file.

        Returns
        -------
        :class:`requests.models.Response`

        Examples
        --------

        Generating and downloading the generated document:

        .. code-block:: python

            import datarobot as dr

            dr.Client(token=my_token, endpoint=endpoint)

            doc = dr.AutomatedDocument(
                    document_type="AUTOPILOT_SUMMARY",
                    entity_id="6050d07d9da9053ebb002ef7",
                    output_format="docx",
                    filepath="/Users/username/Documents/Project_Report_1.docx"
                    )

            doc.generate()
            doc.download()

        Downloading an earlier generated document when you know the document ID:

        .. code-block:: python

            import datarobot as dr

            dr.Client(token=my_token, endpoint=endpoint)
            doc = dr.AutomatedDocument(id='5e8b6a34d2426053ab9a39ed')
            doc.download()

        Notice that ``filepath`` was not set for this document. In this case, the file is saved
        to the directory from which the script was launched.

        Downloading a document chosen from a list of earlier generated documents:

        .. code-block:: python

            import datarobot as dr

            dr.Client(token=my_token, endpoint=endpoint)

            model_id = "6f5ed3de855962e0a72a96fe"
            docs = dr.AutomatedDocument.list_generated_documents(entity_ids=[model_id])
            doc = docs[0]
            doc.filepath = "/Users/me/Desktop/Recommended_model_doc.docx"
            doc.download()

        """
        if not self.id:
            raise AttributeError(
                "Document ID not provided. Assign it to AutomatedDocument object `id` attribute."
            )
        response = self._client.get(f"{self._path}{self.id}/", stream=True)

        if not self.filepath:
            self.filepath = response.headers["Content-Disposition"].split("=")[-1]

        with open(cast(str, self.filepath), mode="wb") as f:
            for chunk in response.iter_content(chunk_size=1024 * 1024):
                f.write(chunk)

        return cast(Response, response)

    def delete(self) -> Response:
        """
        Delete a document using its ID.

        Returns
        -------
        :class:`requests.models.Response`

        Examples
        --------

        .. code-block:: python

            import datarobot as dr

            dr.Client(token=my_token, endpoint=endpoint)
            doc = dr.AutomatedDocument(id="5e8b6a34d2426053ab9a39ed")
            doc.delete()

        If you don't know the document ID, you can follow the same workflow to get the ID as in
        the examples for the ``AutomatedDocument.download`` method.


        """
        if not self.id:
            raise AttributeError(
                "Provide Document ID to delete a document."
                "Assign it to AutomatedDocument object `id` attribute or pass to class constructor."
            )

        return cast(Response, self._client.delete(f"{self._path}{self.id}/"))

    @classmethod
    def list_generated_documents(
        cls,
        document_types: Optional[List[str]] = None,
        entity_ids: Optional[List[str]] = None,
        output_formats: Optional[List[str]] = None,
        locales: Optional[List[str]] = None,
        offset: Optional[int] = None,
        limit: Optional[int] = None,
    ) -> List["AutomatedDocument"]:
        """
        Get information about all previously generated documents available for your account. The
        information includes document ID and type, ID of the entity it was generated for, time of
        creation, and other information.

        Parameters
        ----------
        document_types : List of str or None
            Query for one or more document types.
        entity_ids : List of str or None
            Query generated documents by one or more entity IDs.
        output_formats : List of str or None
            Query for one or more output formats.
        locales : List of str or None
            Query generated documents by one or more locales.
        offset: int or None
            Number of items to skip. Defaults to 0 if not provided.
        limit: int or None
            Number of items to return, maximum number of items is 1000.

        Returns
        -------
        List of AutomatedDocument objects, where each object contains attributes described in
        :py:class:`AutomatedDocument<datarobot.models.automated_documentation.AutomatedDocument>`

        Examples
        --------

        To get a list of all generated documents:

        .. code-block:: python

            import datarobot as dr

            dr.Client(token=my_token, endpoint=endpoint)
            docs = AutomatedDocument.list_generated_documents()


        To get a list of all ``AUTOPILOT_SUMMARY`` documents:

        .. code-block:: python

            import datarobot as dr

            dr.Client(token=my_token, endpoint=endpoint)
            docs = AutomatedDocument.list_generated_documents(document_types=["AUTOPILOT_SUMMARY"])


        To get a list of 5 recently created automated documents in ``html`` format:

        .. code-block:: python

            import datarobot as dr

            dr.Client(token=my_token, endpoint=endpoint)
            docs = AutomatedDocument.list_generated_documents(output_formats=["html"], limit=5)

        To get a list of automated documents created for specific entities (projects or models):

        .. code-block:: python

            import datarobot as dr

            dr.Client(token=my_token, endpoint=endpoint)
            docs = AutomatedDocument.list_generated_documents(
                entity_ids=["6051d3dbef875eb3be1be036",
                            "6051d3e1fbe65cd7a5f6fde6",
                            "6051d3e7f86c04486c2f9584"]
                )

        Note, that the list of results contains ``AutomatedDocument`` objects, which means that you
        can execute class-related methods on them. Here's how you can list, download, and then
        delete from the server all automated documents related to a certain entity:

        .. code-block:: python

            import datarobot as dr

            dr.Client(token=my_token, endpoint=endpoint)

            ids = ["6051d3dbef875eb3be1be036", "5fe1d3d55cd810ebdb60c517f"]
            docs = AutomatedDocument.list_generated_documents(entity_ids=ids)
            for doc in docs:
                doc.download()
                doc.delete()
        """
        params = {
            "document_type": document_types,
            "entity_id": entity_ids,
            "output_format": output_formats,
            "locale": locales,
            "offset": offset,
            "limit": limit,
        }
        params = {camelize(key): val for key, val in params.items() if val is not None}

        if not limit:
            params["limit"] = cls.DEFAULT_BATCH_SIZE
            return [
                cls.from_server_data(item) for item in unpaginate(cls._path, params, cls._client)
            ]

        items = cls._client.get(cls._path, params=params).json()["data"]
        return [cls.from_server_data(item) for item in items]

    def _model_compliance_initialization_status(self, location: str) -> Tuple[bool, str]:
        response = self._client.get(location)
        data = response.json()
        return data["initialized"], data["status"]

    def __repr__(self) -> str:
        attrs = ", ".join(item[0] + "=" + str(item[1]) for item in self.__dict__.items())
        return f"{self.__class__.__name__}({attrs})"



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from typing import Any, Dict, Iterable, List, Optional, Type, TypeVar, Union

import trafaret as t

from datarobot.client import get_client, staticproperty
from datarobot.utils import from_api

T = TypeVar("T", bound="APIObject")
ServerDataDictType = Dict[str, Any]
ServerDataListType = List[ServerDataDictType]
ServerDataType = Union[ServerDataDictType, ServerDataListType]


class APIObject:  # pylint: disable=missing-class-docstring
    _client = staticproperty(get_client)
    _converter = t.Dict({}).allow_extra("*")

    @classmethod
    def _fields(cls):
        return {k.to_name or k.name for k in cls._converter.keys}

    @classmethod
    def from_data(cls: Type[T], data: ServerDataType) -> T:
        """
        Instantiate an object of this class using a dict.

        Parameters
        ----------
        data : dict
            Correctly snake_cased keys and their values.
        """
        checked = cls._converter.check(data)
        safe_data = cls._filter_data(checked)
        return cls(**safe_data)

    @classmethod
    def from_location(
        cls: Type[T],
        path: str,
        keep_attrs: Optional[List[str]] = None,
        params: Optional[Dict] = None,
    ) -> T:
        server_data = cls._server_data(path, params=params)
        return cls.from_server_data(server_data, keep_attrs=keep_attrs)

    @classmethod
    def from_server_data(
        cls: Type[T],
        data: ServerDataType,
        keep_attrs: Optional[Iterable[str]] = None,
    ) -> T:
        """
        Instantiate an object of this class using the data directly from the server,
        meaning that the keys may have the wrong camel casing

        Parameters
        ----------
        data : dict
            The directly translated dict of JSON from the server. No casing fixes have
            taken place
        keep_attrs : iterable
            List, set or tuple of the dotted namespace notations for attributes to keep within the
            object structure even if their values are None
        """
        case_converted = from_api(data, keep_attrs=keep_attrs)
        return cls.from_data(case_converted)

    @classmethod
    def _filter_data(cls: Type[T], data: ServerDataType) -> ServerDataDictType:
        fields = cls._fields()
        return {key: value for key, value in data.items() if key in fields}

    @classmethod
    def _safe_data(cls, data, do_recursive=False):
        return cls._filter_data(cls._converter.check(from_api(data, do_recursive=do_recursive)))

    @classmethod
    def _server_data(cls, path: str, params: Optional[Dict] = None) -> ServerDataType:
        return cls._client.get(path, params=params).json()



#
# Copyright 2021-2022 DataRobot, Inc. and its affiliates.
#
# All rights reserved.
#
# DataRobot, Inc.
#
# This is proprietary source code of DataRobot, Inc. and its
# affiliates.
#
# Released under the terms of DataRobot Tool and Utility Agreement.
from __future__ import annotations

from typing import Any, Dict, List, Optional, TYPE_CHECKING

import trafaret as t

from datarobot.enums import DEFAULT_MAX_WAIT
from datarobot.models.api_object import APIObject
from datarobot.utils import parse_time
from datarobot.utils.pagination import unpaginate
from datarobot.utils.waiters import wait_for_async_resolution

if TYPE_CHECKING:
    from mypy_extensions import TypedDict

    from datarobot.models.api_object import ServerDataDictType

    class SegmentationData(TypedDict):
        datetime_partition_column: Optional[str]
        multiseries_id_columns: Optional[List[str]]
        user_defined_segment_id_columns: Optional[List[str]]

    class SegmentationTaskCreatedResponse(TypedDict):
        numberOfJobs: int
        completedJobs: List[SegmentationTask]
        failedJobs: List[Dict[str, Any]]


class SegmentationTask(APIObject):
    """A Segmentation Task is used for segmenting an existing project into multiple child
    projects. Each child project (or segment) will be a separate autopilot run. Currently
    only user defined segmentation is supported.

    Example for creating a new SegmentationTask for Time Series segmentation with a
    user defined id column:

    .. highlight:: python
    .. code-block:: python

        from datarobot import SegmentationTask

        # Create the SegmentationTask
        segmentation_task_results = SegmentationTask.create(
            project_id=project.id,
            target=target,
            use_time_series=True,
            datetime_partition_column=datetime_partition_column,
            multiseries_id_columns=[multiseries_id_column],
            user_defined_segment_id_columns=[user_defined_segment_id_column]
        )

        # Retrieve the completed SegmentationTask object from the job results
        segmentation_task = segmentation_task_results['completedJobs'][0]

    Attributes
    ----------
    id : ObjectId
        The id of the segmentation task.
    project_id : ObjectId
        The associated id of the parent project.
    type : str
        What type of job the segmentation task is associated with, e.g. auto_ml or auto_ts.
    created  : datetime
        The date this segmentation task was created.
    segments_count : int
        The number of segments the segmentation task generated.
    segments : list of strings
        The segment names that the segmentation task generated.
    metadata : dict
        List of features that help to identify the parameters used by the segmentation task.
    data : dict
        Optional parameters that are associated with enabled metadata for the segmentation task.
    """

    _base_url = "projects/{}/segmentationTasks/"
    _get_url = "projects/{}/segmentationTasks/{}/"
    _results_url = "projects/{}/segmentationTasks/results/{}/"

    _attributes = [
        "id",
        "project_id",
        "name",
        "type",
        "created",
        "segments_count",
        "segments",
        "metadata",
        "data",
    ]

    _converter = t.Dict(
        {
            t.Key("id"): t.String,
            t.Key("project_id"): t.String,
            t.Key("name"): t.String,
            t.Key("type"): t.String,
            t.Key("created"): parse_time,
            t.Key("segments_count"): t.Int,
            t.Key("segments"): t.List(t.String),
            t.Key("metadata"): t.Dict(
                {
                    t.Key("use_time_series"): t.Bool,
                    t.Key("use_multiseries_id_columns"): t.Bool,
                    t.Key("use_automated_segmentation"): t.Bool,
                }
            ).ignore_extra("*"),
            t.Key("data"): t.Dict(
                {
                    t.Key("datetime_partition_column", optional=True): t.Or(t.String, t.Null),
                    t.Key("multiseries_id_columns", optional=True): t.Or(t.List(t.String), t.Null),
                    t.Key("user_defined_segment_id_columns", optional=True): t.Or(
                        t.List(t.String), t.Null
                    ),
                    t.Key("model_package_id", optional=True): t.Or(t.String, t.Null),
                }
            ).ignore_extra("*"),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        id: str,
        project_id: str,
        name: str,
        type: str,
        created: str,
        segments_count: int,
        segments: List[str],
        metadata: Dict[str, bool],
        data: SegmentationData,
    ) -> None:
        # pylint: disable=invalid-name
        self.id = id
        self.project_id = project_id
        self.name = name
        self.type = type
        self.created = created
        self.segments_count = segments_count
        self.segments = segments
        self.metadata = metadata
        self.data = data

    @classmethod
    def from_data(cls, data: ServerDataDictType) -> SegmentationTask:  # type: ignore[override]
        data["id"] = data.get("segmentation_task_id")
        checked = cls._converter.check(data)
        safe_data = cls._filter_data(checked)
        return cls(**safe_data)

    def collect_payload(self) -> Dict[str, str]:
        """Convert the record to a dictionary"""

        out = {attr: getattr(self, attr) for attr in self._attributes}
        return out

    @classmethod
    def create(
        cls,
        project_id: str,
        target: str,
        use_time_series: bool = False,
        datetime_partition_column: Optional[str] = None,
        multiseries_id_columns: Optional[List[str]] = None,
        user_defined_segment_id_columns: Optional[List[str]] = None,
        max_wait: int = DEFAULT_MAX_WAIT,
        model_package_id: Optional[str] = None,
    ) -> SegmentationTaskCreatedResponse:
        """
        Creates segmentation tasks for the project based on the defined parameters.

        Parameters
        ----------
        project_id : str
            The associated id of the parent project.
        target : str
            The column that represents the target in the dataset.
        use_time_series : bool
            Whether AutoTS or AutoML segmentations should be generated.
        datetime_partition_column : str or null
            Required for Time Series.
            The name of the column whose values as dates are used to assign a row
            to a particular partition.
        multiseries_id_columns : list of str or null
            Required for Time Series.
            A list of the names of multiseries id columns to define series within the training
            data. Currently only one multiseries id column is supported.
        user_defined_segment_id_columns : list of str or null
            Required when using a column for segmentation.
            A list of the segment id columns to use to define what columns are used to manually
            segment data. Currently only one user defined segment id column is supported.
        model_package_id : str
            Required when using automated segmentation.
            The associated id of the model in the DataRobot Model Registry that will be used to
            perform automated segmentation on a dataset.
        max_wait : integer
            The number of seconds to wait

        Returns
        -------
        segmentation_tasks : dict
            Dictionary containing the numberOfJobs, completedJobs, and failedJobs. completedJobs
            is a list of SegmentationTask objects, while failed jobs is a list of dictionaries
            indicating problems with submitted tasks.
        """
        payload = {
            "target": target,
            "use_time_series": use_time_series,
        }
        if use_time_series:
            if datetime_partition_column is None:
                raise ValueError(
                    "A datetime_partition_column value must be specified for time series."
                )
            payload.update({"datetime_partition_column": datetime_partition_column})

            if multiseries_id_columns is None:
                raise ValueError(
                    "A multiseries_id_columns value must be specified for time series."
                )
            if not isinstance(multiseries_id_columns, (list, tuple)):
                raise ValueError(
                    "Expected list of str for multiseries_id_columns, got: {}".format(
                        multiseries_id_columns
                    )
                )
            payload.update({"multiseries_id_columns": multiseries_id_columns})

        if user_defined_segment_id_columns is not None:
            if not isinstance(user_defined_segment_id_columns, (list, tuple)):
                raise ValueError(
                    "Expected list of str for user_defined_segment_id_columns, got: {}".format(
                        user_defined_segment_id_columns
                    )
                )
            payload.update({"user_defined_segment_id_columns": user_defined_segment_id_columns})
            payload.update({"use_automated_segmentation": False})
        elif model_package_id is not None:
            payload.update({"model_package_id": model_package_id})
            payload.update({"use_automated_segmentation": True})
        else:
            raise ValueError(
                "Either a user_defined_segment_id_columns value or a model_package_id must "
                "be defined to create a new segmentation task."
            )

        response = cls._client.post(cls._base_url.format(project_id), data=payload)
        results_url = wait_for_async_resolution(
            cls._client, response.headers["Location"], max_wait=max_wait
        )
        results_response = cls._client.get(results_url)

        data = results_response.json()
        successful_jobs = []
        if len(data["completedJobs"]) > 0:
            successful_jobs = [cls.from_location(x["url"]) for x in data["completedJobs"]]

        return {
            "numberOfJobs": data["numberOfJobs"],
            "completedJobs": successful_jobs,
            "failedJobs": data["failedJobs"],
        }

    @classmethod
    def list(cls, project_id: str) -> List[SegmentationTask]:
        """
        List all of the segmentation tasks that have been created for a specific project_id.

        Parameters
        ----------
        project_id : str
            The id of the parent project

        Returns
        -------
        segmentation_tasks : list of SegmentationTask
            List of instances with initialized data.
        """

        return [
            cls.from_server_data(x)
            for x in unpaginate(
                initial_url=cls._base_url.format(project_id),
                initial_params=None,
                client=cls._client,
            )
        ]

    @classmethod
    def get(cls, project_id: str, segmentation_task_id: str) -> SegmentationTask:
        """
        Retrieve information for a single segmentation task associated with a project_id.

        Parameters
        ----------
        project_id : str
            The id of the parent project
        segmentation_task_id : str
            The id of the segmentation task

        Returns
        -------
        segmentation_task : SegmentationTask
            Instance with initialized data.
        """

        return cls.from_location(cls._get_url.format(project_id, segmentation_task_id))


class SegmentInfo(APIObject):
    """
    A SegmentInfo is an object containing information about the combined model segments

    Attributes
    ----------
    project_id : str
        The associated id of the child project.
    segment : str
        the name of the segment
    project_stage : str
        A description of the current stage of the project
    project_status_error : str
        Project status error message.
    autopilot_done : bool
        Is autopilot done for the project.
    model_count : int
        Count of trained models in project.
    model_id : str
        ID of segment champion model.
    """

    _base_url = "projects/{}/combinedModels/{}/segments/"

    _converter = t.Dict(
        {
            t.Key("project_id"): t.String,
            t.Key("segment"): t.String,
            t.Key("project_stage"): t.String,
            t.Key("project_status_error"): t.String(allow_blank=True),
            t.Key("autopilot_done"): t.Bool,
            t.Key("model_count", optional=True): t.Or(t.Int, t.Null),
            t.Key("model_id", optional=True): t.Or(t.String, t.Null),
        }
    ).ignore_extra("*")

    def __init__(
        self,
        project_id: str,
        segment: str,
        project_stage: str,
        project_status_error: str,
        autopilot_done: bool,
        model_count: Optional[int] = None,
        model_id: Optional[str] = None,
    ) -> None:
        self.project_id = project_id
        self.segment = segment
        self.project_stage = project_stage
        self.project_status_error = project_status_error
        self.autopilot_done = autopilot_done
        self.model_count = model_count
        self.model_id = model_id

    def __repr__(self) -> str:
        return "SegmentInfo(segment={}, project_id={}, autopilot_done={})".format(
            self.segment, self.project_id, self.autopilot_done
        )

    @classmethod
    def list(cls, project_id: str, model_id: str) -> List[SegmentInfo]:
        """
        List all of the segments that have been created for a specific project_id.

        Parameters
        ----------
        project_id : str
            The id of the parent project

        Returns
        -------
        segments : list of datarobot.models.segmentation.SegmentInfo
            List of instances with initialized data.
        """

        return [
            cls.from_server_data(x)
            for x in unpaginate(
                initial_url=cls._base_url.format(project_id, model_id),
                initial_params=None,
                client=cls._client,
            )
        ]
